
/**********************************************************************************************/
/*                                                                                            */
/**********************************************************************************************/

#include <upc.h>
#include <stdio.h>
#include <stdlib.h>
#include "upc_task.h"
#include "timers.h"

int mcutoff;

/* global shared task queue handle */
taskq_t *taskq;

/* timers */
static int UPC_RUN=0;

/* function declaration */
void fib(int *np, long *out);
long fib_seq (int n);

static long solutions[] = {
  0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55,    /* 10 */
  89, 144, 233, 377, 610, 987, 1597, 2584, 4181, 6765,    /* 20 */
  10946, 17711, 28657, 46368, 75025, 121393, 196418, 317811, 514229, 832040,    /* 30 */
  1346269, 2178309, 3524578, 5702887, 9227465,                      /* 35 */
  14930352, 24157817, 39088169, 63245986, 102334155,                /* 40 */
  165580141, 267914296, 433494437, 701408733, 1134903170,           /* 45 */
  1836311903, 2971215073, 4807526976, 7778742049, 12586269025,      /* 50 */
  20365011074, 32951280099, 53316291173, 86267571272, 139583862445,  /* 55 */
	225851433717, 365435296162 /* 57 */
};

#define MAX_SOLUTIONS (sizeof(solutions)/sizeof(long) - 1)    /* fib starting from 0 */

long fib_seq (int n)
{
	long x, y;
	int n1 = n-1;
	int n2 = n-2;
	if (n < 2) return n;

	x = fib_seq(n1);
	y = fib_seq(n2);

	return x + y;
}

void fib(int *np, long *out)
{
	int n = *np;
	int n1 = n-1;
	int n2 = n-2;
	long x, y;

	DEBUG3( fprintf(dbg_fp, "\nFIB(%d) -------------------------------------------------- \n", n); )

	if (n < mcutoff) {
		*out = fib_seq(n);
		DEBUG3( fprintf(dbg_fp, "FIB(%d): NOTIFY0: *out=%ld, out=%p\n", n, *out, out);)
	}
	else {
		taskq_put(taskq, fib, &n1, &x);
		taskq_put(taskq, fib, &n2, &y);
		taskq_wait(taskq);
		*out = x + y;
		DEBUG3( fprintf(dbg_fp, "FIB(%d): NOTIFY1: x=%ld, y=%ld\n", n, x, y);)
		DEBUG3( fprintf(dbg_fp, "FIB(%d): NOTIFY1: &x=%p, &y=%p\n", n, &x, &y);)
		DEBUG3( fprintf(dbg_fp, "FIB(%d): NOTIFY1: out=%p, *out=%ld\n", n, out, *out);)
	}
}

int main(int argc, char *argv[]) {
	int i, j, N;
	long result, solution;
	int total_put_count; 
	double total_put_time;
	int steal_size=1; 
	int chunk_size=1; 
	int threshold=2;
	int steal_trial=8;
	int chunk_mode = 1;

	if (argc == 3) {
		N = (int) atoi(argv[1]);
		mcutoff = (int) atoi(argv[2]);
	}
	else if (argc == 2) {
		N = (int) atoi(argv[1]);
		mcutoff = (int) (N/2);
		if (MYTHREAD==0) {
			printf("Note: manual cutoff value is set to %d (N/2) by default\n", mcutoff);
			printf("If you want to change the manual cutoff value, provide one as follows:\n");
			printf("USAGE: upcrun -n <nprocs> fib <N> <manual cutoff>  (e.g. upcrun -n 8 fib 45 20)\n"); 
		}
	}
	else {
		printf("USAGE: upcrun -n <nprocs> fib <N> <manual cutoff>  (e.g. upcrun -n 8 fib 45 20)\n"); 
		upc_global_exit(1);
	}

	timer_clear(UPC_RUN);

	if (MYTHREAD==0) { 
		printf("fib(%d): steal_size=%d chunk_size=%d threshold=%d mcutoff=%d chunk_mode=%d\n", 
						N, steal_size, chunk_size, threshold, mcutoff, chunk_mode);
	}

	taskq = taskq_all_alloc(1, fib, sizeof(int), sizeof(long));

	taskq_set_steal_size(taskq, steal_size);
	taskq_set_chunk_size(taskq, chunk_size);
	taskq_set_threshold(taskq, threshold);

  /* - random - */
	taskq_set_steal_trial(taskq, 0, steal_trial);
  taskq_set_steal_mode(taskq, FULL_RAND_STEAL, 1, THREADS);
	taskq_set_chunk_mode(taskq, chunk_mode);

	upc_barrier;

	/* ----------------- */

	timer_start(UPC_RUN);

	if (MYTHREAD==0) {
		taskq_put(taskq, fib, &N, &result);
	}

	taskq_all_run(taskq);

	timer_stop(UPC_RUN);

	fprintf(dbg_fp, "FINAL: result=%ld, &result=%p\n", result, &result);

	/* ----------------- */

	upc_barrier;

	taskq_check(taskq);
	taskq_stats(taskq);

	if (MYTHREAD==0) {

		printf("Final result = %ld\n", result);
		
		if (N > MAX_SOLUTIONS) printf("SOLUTION not available for N=%d\n", N);
		else {
			solution = solutions[N];
			if (solution == result)
				printf("OOOOO Correct Fibbonachi(%d) result=%ld\n", N, result);
			else
				printf("XXXXX Incorrect Fibbonachi(%d) result=%ld, solution=%ld\n", N, result, solution);
		}

		printf("UPC_RUN=%lf sec\n", timer_read(UPC_RUN));

	}

	upc_barrier;

	return 0;
}

