#include <stdio.h>
#include <stdint.h>
#include <stdlib.h> 
#include <string.h>
#include <math.h>
#include <libgen.h>
#include <upc.h>
#include "upc_task.h"
#include "timers.h"

#define EPSILON 1.0E-6
#define B 2	

#define RESULT_SUCCESSFUL 1
#define RESULT_UNSUCCESSFUL 2

//#define HOPPER

#define VERIFY		/* do not turn on VERIFY, it may affect overall performance */	

//#define TASK_TIMER	// comment out for perf. evaluation

/* timers */
static int UPC_RUN=0;
static int SER_RUN=1;
#ifdef TASK_TIMER
static int FWD_TIME = 2;
static int BDIV_TIME = 3;
static int BMOD_TIME = 4;
#endif

int	N1, N2;
float *inner_buffer, *row_buffer, *col_buffer, *diag_buffer;

taskq_t *taskq;

int num_clean_blocks;
int clean_block_count = 0;
shared [] float *clean_block_sp;

typedef struct data2 data2_t;
struct data2{
	shared [] float *x;
	shared [] float *y;
};

typedef struct data3 data3_t;
struct data3{
	shared [] float *x;
	shared [] float *y;
	shared [] float *z;
};

/*----------------------------------------------------------------------
 * bmod_core function is outlined to be separated from the bmod_task 
 * function. If it is not outlined, it may cause some compilers not to 
 * fully optimize this triple-nested loop 
 *---------------------------------------------------------------------- */
void bmod_core(int N2, float *inner_pp, float *row_pp, float *col_pp)
{
  int i, j, k;

	for (i=0; i<N2; i++)
		for (j=0; j<N2; j++)
			for (k=0; k<N2; k++)
				inner_pp[i*N2+j] = inner_pp[i*N2+j] - row_pp[i*N2+k]*col_pp[k*N2+j];
}

/***********************************************************************
 * checkmat: 
 **********************************************************************/
int par_checkmat (float *M, shared [] float *N)
{
   int i, j;
   float r_err;

   for (i = 0; i < N2; i++) 
   {
      for (j = 0; j < N2; j++) 
      {
         r_err = M[i*N2+j] - N[i*N2+j];
         if (r_err < 0.0 ) r_err = -r_err;
         r_err = r_err / M[i*N2+j];
         if(r_err > EPSILON)
         {
            fprintf(dbg_fp, "Checking failure: A[%d][%d]=%f  B[%d][%d]=%f; Relative Error=%f\n",
                    i,j, M[i*N2+j], i,j, N[i*N2+j], r_err);
            return FALSE;
         }
      }
   }
   return TRUE;
}

void print_owner(shared [] float * shared [B] *M)
{
	int ii, jj;

  fprintf(dbg_fp, "---------------- Print Owner ----------------\n");
	for (ii=0; ii<N1; ii++) {
		for (jj=0; jj<N1; jj++) {
			fprintf(dbg_fp, "upc_threadof(&M[%d]) = %lu ", ii*N1+jj, upc_threadof(&M[ii*N1+jj]) ); 
			fprintf(dbg_fp, "upc_threadof(M[%d]) = %lu\n", ii*N1+jj, upc_threadof(M[ii*N1+jj]) ); 
		}
	}
}

#ifdef VERIFY
void print_mat(float *p)
{
	int i, j;
  fprintf(dbg_fp, "------------- Print Matrix (SEQ) -------------\n");
	for (i=0; i<N2; i++) {
		fprintf(dbg_fp, "m[%d, :] = ", i);
		for (j=0; j<N2; j++) {
			fprintf(dbg_fp, "%10.6f ", p[i*N2+j]);
		}
		fprintf(dbg_fp, "\n");
	}
  fprintf(dbg_fp, "\n----------------------------------------------\n");
}

void par_print_mat(shared [] float *p)
{
	int i, j;

  fprintf(dbg_fp, "------------- Print Matrix (PAR) -------------\n");
	for (i=0; i<N2; i++) {
		fprintf(dbg_fp, "m[%d, :] = ", i);
		for (j=0; j<N2; j++) {
			fprintf(dbg_fp, "%10.6f ", p[i*N2+j]);
		}
		fprintf(dbg_fp, "\n");
	}
  fprintf(dbg_fp, "\n----------------------------------------------\n");
}

void print_all_mat(float * M[])
{
  int ii, jj;
	int i, j;
  fprintf(dbg_fp, "---------------- Print Matrix (SEQ) ----------------\n");
  for (ii=0; ii < N1; ii++) {
    for (jj=0; jj < N1; jj++) {
    	fprintf(dbg_fp, "M[%d, %d]: ", ii, jj);
      if (M[ii*N1+jj]==NULL)
        fprintf(dbg_fp, "  NULL\n");
      else {
				print_mat( M[ii*N1+jj] );
			}
    }
    fprintf(dbg_fp, "\n");
  }
  fprintf(dbg_fp, "\n----------------------------------------------\n");
}

void par_print_all_mat(shared [] float * shared [B] M[])
{
  int ii, jj;
	int i, j;
  fprintf(dbg_fp, "---------------- Print Matrix (PAR) ----------------\n");
  for (ii=0; ii < N1; ii++) {
    for (jj=0; jj < N1; jj++) {
    	fprintf(dbg_fp, "M[%d, %d]: ", ii, jj);
      if (M[ii*N1+jj]==NULL)
        fprintf(dbg_fp, "  NULL\n");
      else {
				par_print_mat( M[ii*N1+jj] );
			}
    }
    fprintf(dbg_fp, "\n");
  }
  fprintf(dbg_fp, "\n----------------------------------------------\n");
}
#endif

/***********************************************************************
 * genmat: 
 **********************************************************************/
#ifdef VERIFY
void genmat (float *M[])
{
  int null_entry, init_val, i, j, ii, jj;
  float *p;
  int a=0,b=0;

  init_val = 1325;

  /* generating the structure */
  for (ii=0; ii < N1; ii++)
  {
 		for (jj=0; jj < N1; jj++)
    {
      /* computing null entries */
      null_entry=FALSE;
      if ((ii<jj) && (ii%3 !=0)) null_entry = TRUE;
      if ((ii>jj) && (jj%3 !=0)) null_entry = TRUE;
	 		if (ii%2==1) null_entry = TRUE;
	 		if (jj%2==1) null_entry = TRUE;
	 		if (ii==jj) null_entry = FALSE;
	 		if (ii==jj-1) null_entry = FALSE;
      if (ii-1 == jj) null_entry = FALSE; 
      /* allocating matrix */
      if (null_entry == FALSE){
        a++;
        M[ii*N1+jj] = (float *) malloc(N2*N2*sizeof(float));
	  		if ((M[ii*N1+jj] == NULL))
        {
          fprintf(dbg_fp, "Error: Out of memory\n");
          exit(101);
        }
        /* initializing matrix */
        p = M[ii*N1+jj];
        for (i = 0; i < N2; i++) 
        {
          for (j = 0; j < N2; j++)
          {
	          init_val = (3125 * init_val) % 65536;
      	    (*p) = (float)((init_val - 32768.0) / 16384.0);
            p++;
          }
        }
      }
      else
      {
        b++;
        M[ii*N1+jj] = NULL;
      }
    }
  }
  fprintf(dbg_fp, "SER: allo=%d, no=%d, total=%d, factor=%f\n",a,b,a+b,(float)((float)a/(float)(a+b)));
}
#endif

void par_genmat (shared [] float * shared [B] *M)
{
  int null_entry, init_val, i, j, ii, jj, index;
  shared [] float *p;
  int a=0,b=0;
	int write_flag = 0;

  init_val = 1325;

  /* generating the structure */
  for (ii=0; ii < N1; ii++)
  {
 		for (jj=0; jj < N1; jj++)
    {
			if ( upc_threadof( &M[ii*N1+jj] ) == MYTHREAD ) {
				write_flag = 1;
			}
			else {
				write_flag = 0;
			}
	      /* computing null entries */
	      null_entry=FALSE;
	      if ((ii<jj) && (ii%3 !=0)) null_entry = TRUE;
	      if ((ii>jj) && (jj%3 !=0)) null_entry = TRUE;
		 		if (ii%2==1) null_entry = TRUE;
		 		if (jj%2==1) null_entry = TRUE;
		 		if (ii==jj) null_entry = FALSE;
		 		if (ii==jj-1) null_entry = FALSE;
	      if (ii-1 == jj) null_entry = FALSE; 

	      /* allocating matrix */
	      if (null_entry == FALSE){
	        a++;
					if (write_flag) {
	        	M[ii*N1+jj] = (shared [] float *) upc_alloc(N2*N2*sizeof(float));
		  			if ((M[ii*N1+jj] == NULL))
		        {
		          fprintf(dbg_fp, "Error: Out of memory\n");
		          exit(101);
		        }
					}
	        /* initializing matrix */
	        p = M[ii*N1+jj];
	        for (i = 0; i < N2; i++) 
	        {
	          for (j = 0; j < N2; j++)
	          {
		          init_val = (3125 * init_val) % 65536;
							if (write_flag==1) {
								p[i*N2+j] = (float)((init_val - 32768.0) / 16384.0);
							}
	          }
	        }
	      }
	      else
	      {
	        b++;
					if (write_flag==1) {
						M[ii*N1+jj] = NULL;
					}
	      }
    }
  }
  fprintf(dbg_fp, "UPC: allo=%d, no=%d, total=%d, factor=%f\n",a,b,a+b,(float)((float)a/(float)(a+b)));
}

/***********************************************************************
 * print_structure: 
 **********************************************************************/
void par_print_structure(char *name, shared [] float * shared [B] M[])
{
	int ii, jj; 
	int min = 2147483647, max = 0;
	int *count = (int *)malloc(THREADS*sizeof(int));
	
	if (MYTHREAD==0) for (ii=0; ii<THREADS; ii++) count[THREADS] = 0;

	fprintf(dbg_fp, "Structure for matrix %s @ 0x%p\n",name, (float **)M);
	for (ii = 0; ii < N1; ii++) {
		for (jj = 0; jj < N1; jj++) {
			if (M[ii*N1+jj]!=NULL) {
				fprintf(dbg_fp, "%lu", upc_threadof(&M[ii*N1+jj]));
				count[upc_threadof(&M[ii*N1+jj])]++;
			}
			else fprintf(dbg_fp, " ");
		}
		fprintf(dbg_fp, "\n");
	}
	fprintf(dbg_fp, "\n");

	if (MYTHREAD==0) {
		printf("Number of the main N1xN1 array element that belongs to each thread:");
		for (ii=0; ii<THREADS; ii++) {
			if ((ii%10)==0) printf("\n");
			printf("T%d=%d ", ii, count[ii]);
			if (count[ii] > max) max = count[ii];
			if (count[ii] < min) min = count[ii];
		}
		printf("\nmin=%d, max=%d\n\n", min, max);
	}
	free(count);
}

/***********************************************************************
 * allocate_clean_block: 
 **********************************************************************/
#ifdef VERIFY
float * allocate_clean_block()
{
  int i,j;
  float *p, *q;

  p = (float *) malloc(N2*N2*sizeof(float));

  q=p;
  if (p!=NULL){
     for (i = 0; i < N2; i++) 
        for (j = 0; j < N2; j++){(*p)=0.0; p++;}
	
  }
  else
  {
      fprintf(dbg_fp, "Error: Out of memory\n");
      exit (101);
  }
  return (q);
}
#endif

shared [] float *pre_allocate_clean_blocks() {

/*
	num_clean_blocks = (N1 * (1 << ((ilog2(N1)-4))));
	num_clean_blocks *= 2; // pre-cautionary work
*/

	num_clean_blocks = (N1*(N1-1))/THREADS;  /* original: (N1*(N1-1)*2)/THREADS; */

	if (MYTHREAD==0) printf("num_clean_blocks for thread 0 = %d\n", num_clean_blocks);

	return (shared [] float *) upc_alloc( num_clean_blocks * N2*N2*sizeof(float) );
}

shared [] float * par_allocate_clean_block(int ii, int jj)
{
  int i, j, index;
  shared [] float *p, *q;

/*
  p = (shared [] float *) upc_alloc(N2*N2*sizeof(float));
*/
	p = (shared [] float *) &clean_block_sp[clean_block_count * N2*N2];

	if (++clean_block_count >= num_clean_blocks) {
		printf("ERROR: tid%d clean block overflow! ", MYTHREAD);
		printf("(%d >= %d)\n", clean_block_count, num_clean_blocks);
		upc_global_exit(1);
	}

  q=p;
  if (p!=NULL){
     for (i = 0; i < N2; i++) 
        for (j = 0; j < N2; j++){(*p)=0.0; p++;}
	
  }
  else
  {
      fprintf(dbg_fp, "Error: Out of memory\n");
      exit (101);
  }
  return (q);
}

/***********************************************************************
 * lu0: 
 **********************************************************************/
#ifdef VERIFY
void lu0(float *diag)
{
   int i, j, k;

   for (k=0; k<N2; k++)
      for (i=k+1; i<N2; i++)
      {
         diag[i*N2+k] = diag[i*N2+k] / diag[k*N2+k];
         for (j=k+1; j<N2; j++)
            diag[i*N2+j] = diag[i*N2+j] - diag[i*N2+k] * diag[k*N2+j];
      }
}
#endif

void par_lu0(shared [] float *diag)
{
	int i, j, k;

	for (k=0; k<N2; k++) {
		for (i=k+1; i<N2; i++) {
			diag[i*N2+k] = diag[i*N2+k] / diag[k*N2+k];
			for (j=k+1; j<N2; j++) {
				diag[i*N2+j] = diag[i*N2+j] - diag[i*N2+k] * diag[k*N2+j];
			}
		}
	}
}

/***********************************************************************
 * bdiv: 
 **********************************************************************/
#ifdef VERIFY
void bdiv(float *diag, float *row)
{
   int i, j, k;
   for (i=0; i<N2; i++)
      for (k=0; k<N2; k++)
      {
         row[i*N2+k] = row[i*N2+k] / diag[k*N2+k];
         for (j=k+1; j<N2; j++)
            row[i*N2+j] = row[i*N2+j] - row[i*N2+k]*diag[k*N2+j];
      }
}
#endif

void bdiv_task(data2_t *idata2)
{
	int i, j, k;
#ifdef TASK_TIMER
	timer_start(BDIV_TIME);
#endif
	int size = N2*N2*sizeof(float);
	shared [] float *diag = idata2->x;
	shared [] float *row = idata2->y;
	float *diag_pp = diag_buffer; 
	float *row_pp = row_buffer;

	upc_memget(diag_pp, diag, size); 
	upc_memget(row_pp, row, size); 

	for (i=0; i<N2; i++) {
		for (k=0; k<N2; k++)
		{
			row_pp[i*N2+k] = row_pp[i*N2+k] / diag_pp[k*N2+k];
			for (j=k+1; j<N2; j++)
				row_pp[i*N2+j] = row_pp[i*N2+j] - row_pp[i*N2+k]*diag_pp[k*N2+j];
		}
	}

	upc_memput(row, row_pp, size);

#ifdef TASK_TIMER
	timer_stop(BDIV_TIME);
#endif

}

/***********************************************************************
 * bmod: 
 **********************************************************************/
#ifdef VERIFY
void bmod(float *row, float *col, float *inner)
{
   int i, j, k;
   for (i=0; i<N2; i++)
      for (j=0; j<N2; j++)
         for (k=0; k<N2; k++)
            inner[i*N2+j] = inner[i*N2+j] - row[i*N2+k]*col[k*N2+j];
}
#endif

void bmod_task(data3_t *idata3)
{
  int i, j, k;
#ifdef TASK_TIMER
	timer_start(BMOD_TIME);
#endif
	int size = (N2*N2*sizeof(float));
	shared [] float *row = idata3->x;
	shared [] float *col = idata3->y;
	shared [] float *inner = idata3->z;
	float *inner_pp = inner_buffer;
	float *row_pp = row_buffer;
	float *col_pp = col_buffer;

	upc_memget(inner_pp, inner, size); 
	upc_memget(row_pp, row, size); 
	upc_memget(col_pp, col, size); 

	bmod_core(N2, inner_pp, row_pp, col_pp);
/*
	for (i=0; i<N2; i++)
		for (j=0; j<N2; j++)
			for (k=0; k<N2; k++)
				inner_pp[i*N2+j] = inner_pp[i*N2+j] - row_pp[i*N2+k]*col_pp[k*N2+j];
*/

	upc_memput(inner, inner_pp, size);
#ifdef TASK_TIMER
	timer_stop(BMOD_TIME);
#endif
}

/***********************************************************************
 * fwd: 
 **********************************************************************/
#ifdef VERIFY
void fwd(float *diag, float *col)
{
   int i, j, k;
   for (j=0; j<N2; j++)
      for (k=0; k<N2; k++) 
         for (i=k+1; i<N2; i++)
            col[i*N2+j] = col[i*N2+j] - diag[i*N2+k]*col[k*N2+j];
}
#endif

void fwd_task(data2_t *idata2)
{
  int i, j, k;
#ifdef TASK_TIMER
	timer_start(FWD_TIME);
#endif
	int size = (N2*N2*sizeof(float));
	shared [] float *diag = idata2->x;
	shared [] float *col = idata2->y;
	float *diag_pp = diag_buffer;
	float *col_pp = col_buffer;

	upc_memget(diag_pp, diag, size); 
	upc_memget(col_pp, col, size); 

	for (j=0; j<N2; j++)
		for (k=0; k<N2; k++) 
			for (i=k+1; i<N2; i++)
				col_pp[i*N2+j] = col_pp[i*N2+j] - diag_pp[i*N2+k]*col_pp[k*N2+j];

	/* upc_memput will write back more data that this routine hasn't written */
	upc_memput(col, col_pp, size);
#ifdef TASK_TIMER
	timer_stop(FWD_TIME);
#endif
}

#ifdef VERIFY
void sparselu_seq_init (float ***pBENCH, char *pass)
{
	*pBENCH = (float **) malloc(N1*N1*sizeof(float *));
  genmat(*pBENCH);
  // print_structure(pass, *pBENCH);
}
#endif

#ifdef VERIFY
void sparselu_seq_call(float **BENCH)
{
   int ii, jj, kk;

   for (kk=0; kk<N1; kk++)
   {
      lu0(BENCH[kk*N1+kk]);
      for (jj=kk+1; jj<N1; jj++)
         if (BENCH[kk*N1+jj] != NULL)
         {
            fwd(BENCH[kk*N1+kk], BENCH[kk*N1+jj]);
         }
      for (ii=kk+1; ii<N1; ii++) 
         if (BENCH[ii*N1+kk] != NULL)
         {
            bdiv (BENCH[kk*N1+kk], BENCH[ii*N1+kk]);
         }
      for (ii=kk+1; ii<N1; ii++)
         if (BENCH[ii*N1+kk] != NULL)
            for (jj=kk+1; jj<N1; jj++)
               if (BENCH[kk*N1+jj] != NULL)
               {
                     if (BENCH[ii*N1+jj]==NULL) BENCH[ii*N1+jj] = allocate_clean_block();
                     bmod(BENCH[ii*N1+kk], BENCH[kk*N1+jj], BENCH[ii*N1+jj]);
               }

   }
}
#endif

void sparselu(shared [] float * shared [B] *BENCH)
{
	int ii, jj, kk;
	data2_t idata2;   
	data3_t idata3;   
#ifdef TASK_TIMER
	double t1 = get_wctime();
#endif

  fprintf(dbg_fp, "Computing SparseLU Factorization (%dx%d matrix with %dx%d blocks)\n", N1,N1,N2,N2);

	inner_buffer = (float *) malloc(N2*N2*sizeof(float));
	diag_buffer = (float *) malloc(N2*N2*sizeof(float));
	row_buffer = (float *) malloc(N2*N2*sizeof(float));
	col_buffer = (float *) malloc(N2*N2*sizeof(float));

  for (kk=0; kk<N1; kk++) 
  {
		if ( upc_threadof( &BENCH[kk*N1+kk] ) == MYTHREAD ) {
    	par_lu0(BENCH[kk*N1+kk]);
		}

		taskq_local_reset(taskq);

		upc_barrier;

		upc_forall (jj=kk+1; jj<N1; jj++; &BENCH[kk*N1+jj]) {
			if (BENCH[kk*N1+jj] != NULL) {
				idata2.x = BENCH[kk*N1+kk];
				idata2.y = BENCH[kk*N1+jj];
				taskq_put(taskq, fwd_task, &idata2);
			}
		}

		taskq_fence(taskq);
		//taskq_wait(taskq);

		upc_forall (ii=kk+1; ii<N1; ii++; &BENCH[ii*N1+kk]) {
			if (BENCH[ii*N1+kk] != NULL) {
				idata2.x = BENCH[kk*N1+kk];
				idata2.y = BENCH[ii*N1+kk];
				taskq_put(taskq, bdiv_task, &idata2);
			}
		}

		taskq_all_run(taskq);

		for (ii=kk+1; ii<N1; ii++) {
			if (BENCH[ii*N1+kk] != NULL) {
				for (jj=kk+1; jj<N1; jj++) {
					if ( upc_threadof( &BENCH[ii*N1+jj] ) == MYTHREAD ) {
						if (BENCH[kk*N1+jj] != NULL) {
							if (BENCH[ii*N1+jj]==NULL) { BENCH[ii*N1+jj] = par_allocate_clean_block(ii, jj); }
							idata3.x = BENCH[ii*N1+kk];
							idata3.y = BENCH[kk*N1+jj];
							idata3.z = BENCH[ii*N1+jj];
							taskq_put(taskq, bmod_task, &idata3);
						}
					}
				}
			}
		}

		taskq_all_run(taskq);
	}

	fprintf(dbg_fp, " completed!\n");
}

#ifdef VERIFY
void sparselu_seq_fini (float **BENCH, char *pass)
{
   // print_structure(pass, BENCH);
}
#endif

#ifdef VERIFY
int par_sparselu_check(float **SEQ, shared [] float * shared [B] *BENCH)
{
   int ii,jj,ok=1;

   for (ii=0; ((ii<N1) && ok); ii++)
   {
      for (jj=0; ((jj<N1) && ok); jj++)
      {
         if ((SEQ[ii*N1+jj] == NULL) && (BENCH[ii*N1+jj] != NULL)) ok = FALSE;
         if ((SEQ[ii*N1+jj] != NULL) && (BENCH[ii*N1+jj] == NULL)) ok = FALSE;
         if ((SEQ[ii*N1+jj] != NULL) && (BENCH[ii*N1+jj] != NULL))
            ok = par_checkmat(SEQ[ii*N1+jj], BENCH[ii*N1+jj]);
				if (!ok) { print_mat(SEQ[ii*N1+jj]); par_print_mat(BENCH[ii*N1+jj]); }
      }
   }
   if (ok) return RESULT_SUCCESSFUL;
   else return RESULT_UNSUCCESSFUL;
}
#endif

int main(int argc, char *argv[])
{
	shared [] float * shared [B] *BENCH;
	float **SEQ; 
	int steal_size, chunk_size, threshold, chunk_mode;

 /* at HOPPER: 
			N1 = 256; N2 = 32; steal_size = 10; chunk_size = 10; threshold = 10; rcutoff = 20; chunk_mode = 1;
	*/

	if (argc != 3) {
		if (MYTHREAD==0) {
			printf("Usage: upcrun -n <num_procs> -shared-heap=1800MB ./sparselu <N1> <N2>\n"); 
		}
		exit(1);
	}

	N1 = (int) atoi(argv[1]);	
	N2 = (int) atoi(argv[2]);	
	steal_size = 10;
	chunk_size = 10;
	threshold = 20;
	chunk_mode = 1;

	if (MYTHREAD==0) printf("Running SparseLU with N1=%d N2=%d\n", N1, N2); 

	if (B > N1) {
		if (MYTHREAD==0) {
			printf("\nERROR: The main matrix size, (N1=%d), cannot be smaller than ", N1);
			printf("the blocking factor, B(=%d).\n       ", B);
			printf("Re-run with larger N1 or re-compile with smaller B.\n       Exiting...\n\n");
		}
		upc_global_exit(1);
	}

	timer_clear(UPC_RUN);
	timer_clear(SER_RUN);
#ifdef TASK_TIMER
	timer_clear(FWD_TIME);
	timer_clear(BDIV_TIME);
	timer_clear(BMOD_TIME);
#endif

	taskq = taskq_all_alloc(3, fwd_task, sizeof(data2_t), 0, bdiv_task, sizeof(data2_t), 0,
													bmod_task, sizeof(data3_t), 0);

  taskq_set_steal_size(taskq, steal_size);
  taskq_set_chunk_size(taskq, chunk_size);
  taskq_set_threshold(taskq, threshold);
	taskq_set_chunk_mode(taskq, chunk_mode);
  taskq_set_independent_task(taskq, bmod_task);

/* HVS: SMP */
	taskq_set_steal_trial(taskq, 0, 12);
	taskq_set_steal_mode(taskq, FULL_RAND_STEAL, 1, THREADS);

/* HVS: Cluster
	taskq_set_steal_trial(taskq, 21, 24);
	taskq_set_steal_mode(taskq, HIER_WORK_STEAL, 2, 8, THREADS);		// HiVS: cluster
*/

	upc_barrier;

#ifdef VERIFY
	if (MYTHREAD==0) printf("Sequential execution starts\n");
	sparselu_seq_init(&SEQ, "benchmark");
	timer_start(SER_RUN);
	sparselu_seq_call(SEQ);
	timer_stop(SER_RUN);
	sparselu_seq_fini(SEQ, "benchmark");
	if (MYTHREAD==0) printf("Sequential execution done\n");
#endif

	// UPC version of sparselu_init
	{
		BENCH = (shared [] float * shared [B] *) upc_all_alloc(N1*N1, sizeof(float *)); 

		clean_block_sp = pre_allocate_clean_blocks();

		par_genmat(BENCH);

		upc_barrier;

		if (MYTHREAD==0) printf("Parallel initialization done\n");

		//par_print_structure("benchmark", BENCH);
	}

	upc_barrier;

	timer_start(UPC_RUN);

	sparselu(BENCH);

	timer_stop(UPC_RUN);

	//par_print_structure("benchmark", BENCH);

	taskq_check(taskq);
	taskq_stats(taskq);

	upc_barrier;

	if (MYTHREAD==0) {
#ifdef VERIFY
		printf("UPC RUN=%lf sec, SER RUN=%lf sec\n", timer_read(UPC_RUN), timer_read(SER_RUN));
		if (par_sparselu_check(SEQ, BENCH) == RESULT_SUCCESSFUL)
			printf("\nVerification Successful!\n");
		else
			printf("\nVerification Failed!\n");
#else
		printf("UPC RUN=%lf sec\n", timer_read(UPC_RUN));
#endif
	}

#ifdef TASK_TIMER
	int fwd_cnt = timer_get_count(FWD_TIME);
	int bdiv_cnt = timer_get_count(BDIV_TIME);
	int bmod_cnt = timer_get_count(BMOD_TIME);
	int total_cnt = fwd_cnt + bdiv_cnt + bmod_cnt;
	double fwd_time = timer_read(FWD_TIME);
	double bdiv_time = timer_read(BDIV_TIME);
	double bmod_time = timer_read(BMOD_TIME);
	double total_time = fwd_time + bdiv_time + bmod_time;
  fprintf(dbg_fp, "fwd_task count=%d total time=%lf sec avg fwd_task_time = %lf\n", 
					fwd_cnt, fwd_time, fwd_time*1000000000/(double)fwd_cnt);
  fprintf(dbg_fp, "bdiv_task count=%d total time=%lf sec avg bdiv_task_time = %lf\n", 
					bdiv_cnt, bdiv_time, bdiv_time*1000000000/(double)bdiv_cnt);
  fprintf(dbg_fp, "bmod_task count=%d total time=%lf sec avg bmod_task_time = %lf\n", 
					bmod_cnt, bmod_time, bmod_time*1000000000/(double)bmod_cnt);
  fprintf(dbg_fp, "all count=%d total time=%lf sec, avg task time = %lf ns\n", 
					total_cnt, total_time, total_time*1000000000/(double)total_cnt);
#endif

}
