#include <stdlib.h>
#include "timers.h"

static double _start_time[MAX_TIMERS];
static double _elapsed_time[MAX_TIMERS];
static int _counter[MAX_TIMERS];

double get_wctime(void) {
#ifdef CLOCK_REALTIME
  //  double t = bupc_ticks_to_us(bupc_ticks_now()) * 1.0e-6;
  struct timespec ts;
  syscall(__NR_clock_gettime, CLOCK_REALTIME, &ts);
  double t = (double)ts.tv_sec + (double)ts.tv_nsec * 1e-9;
  return t;
#else
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (tv.tv_sec + 1E-6 * tv.tv_usec);
#endif
}

void clear_timers() {
	int i;
  for (i = 0; i < MAX_TIMERS; i++) timer_clear(i);
}

void timer_clear(int n) {
  _elapsed_time[n] = 0.0;
  _counter[n] = 0.0;
}

void timer_start(int n) {
  _start_time[n] = get_wctime();
	_counter[n]++;
}

void timer_stop(int n) {
  double t, now;
  now = get_wctime();
  t = now - _start_time[n];
  _elapsed_time[n] += t;
}

double timer_read(int n) {
  return(_elapsed_time[n]);
}

int timer_get_count(int n) {
  return(_counter[n]);
}





