/************************************************
 *****                                      *****
 ***** 	    UPC TASK PARALLEL LIBRARY       *****
 *****                                      *****
 ************************************************/

#include "upc_task.h"

enum STEAL_MODE{ _random=0, _laws_rand=1, _laws_next=2, _intra_node_only=3}; 

#ifdef __BERKELEY_UPC__
/* BUPC nonblocking I/O handles */
bupc_handle_t cb_handle = BUPC_COMPLETE_HANDLE;
const int			local_cb_cancel = 1;
#endif

#define KDEBUG1					// redirects not only out_offs but also optr to the input of the dependent task

#define DEBUG_STACK

//#define DBG_STEAL
#ifdef DBG_STEAL
#define DEBUG_STEAL(stmts) STACK_INDENT; stmts fflush(dbg_fp);
#else
#define DEBUG_STEAL(stmts)
#endif

//#define ENABLE_TIMER
#ifdef ENABLE_TIMER
#define PUT_TIME 48 
#endif

// ----------------------------------------------------------
// GLOBAL VRIABLES
// ----------------------------------------------------------
int *victim_count;

//#define ANALYSIS
#ifdef ANALYSIS
#define BIGNUM 4000000   // UTS benchmark may need more 
//#define SAMPLE_FREQ 10000		// for UTS
#define SAMPLE_FREQ 1
int sample_count1=0;
int sample_count2=0;
int pcnt_top[BIGNUM];
int scnt_top[BIGNUM];
#endif

shared int rcutoff_counts[THREADS];
int	rcutoff_count = 0;	

#ifdef __BERKELEY_UPC__
#define UPC_POLL_NETWORK bupc_poll()
#else
#define UPC_POLL_NETWORK upc_fence
#endif

#define NO_TERM 0
#define TERM 1

/**********************************************************************
 *  Function Declarations                                             *
 **********************************************************************/
int *gen_victim_list(taskq_t *);
//void notify_to_dep_tasks(taskq_t *, task_header_t *, void *);

/**********************************************************************
 *  Utility Function Definition                                       *
 **********************************************************************/

void debug_stack(taskq_handle taskq, const char *fname) {
	int i;
	task_header_t *task_p = &taskq->task_get_buf[0];

	DEBUG3( fprintf(dbg_fp, "%s stack_top=%d : stack taskID = [", fname, taskq->task_stack_top); )
	for (i=0; i<taskq->task_stack_top; i++) {
		task_p =(task_header_t *)((char *)(&taskq->task_get_buf[0]) + i*taskq->task_sz);
		fprintf(dbg_fp, "%lu ", task_p->taskID);
	}
	fprintf(dbg_fp, "]\n");
}

#if 0
void print_stack_indent() {
	for (int i=0; i<g_taskq->task_stack_top; i++) fprintf(dbg_fp, "  ");
}
#endif

#ifdef ANALYSIS
void rtaskq_snapshot(rtaskq_t *rtaskq_lp) {
	if (++sample_count1%SAMPLE_FREQ==0) {
		scnt_top[sample_count2] = rtaskq_lp->scnt;
		pcnt_top[sample_count2] = rtaskq_lp->pcnt;
		if (++sample_count2 == BIGNUM) {
			printf("\n\nWARNING: sample_count overflow! sample_count2=%d\n", sample_count2);
			fprintf(dbg_fp, "\n\nWARNING: sample_count overflow! sample_count2=%d\n", sample_count2);
			for (int i=0; i<sample_count2; i++) {
				fprintf(dbg_fp, "scnt_top = %5d pcnt_top = %5d\n", scnt_top[i], pcnt_top[i]); 
			}
			sample_count1=0;
			sample_count2=0;
		}
	}
}
#endif

int ilog2(int i) {
  int log2;
  int exp2 = 1;
  if (i <= 0) return(-1);
  for (log2 = 0; log2 < 20; log2++) {
    if (exp2 >= i) return(log2);
    exp2 *= 2;
  }
  return(-1);
}

int cbarrier_inc(taskq_t *taskq) {
	//DEBUG3( fprintf(dbg_fp, "[cbarrier_inc] strt: cb_count=%d\n", cb_count); )
	upc_lock(cb_lock);
	{
		cb_count++;
	}
	upc_unlock(cb_lock);
	//DEBUG3( fprintf(dbg_fp, "[cbarrier_inc] done: cb_count=%d\n", cb_count); )

	if (cb_count == THREADS) return TERM;
	else return NO_TERM;
}

int cbarrier_dec(taskq_t *taskq) {
	int status = TERM;

	upc_lock(cb_lock);
	{
		if (cb_count < THREADS) {
			status = NO_TERM;
			cb_count--;
		}
	}
	upc_unlock(cb_lock);

	return status;
}

int snoopWork(taskq_t *taskq) {
	int first_trial = 1;

	if ( taskq_find_victim(taskq, 1, first_trial) != NOT_FOUND )
		return 1;
	else 
		return 0;
}

int detectWork(taskq_t *taskq) {
	if (cb_cancel) {
		DEBUG3( fprintf(dbg_fp, "[detectWork] cb_cancel=%d\n", cb_cancel); )
		return 1;
	}
	else return snoopWork(taskq);
}

void taskq_error_exit(const char *str) {
	printf("tid %d: %s\n", MYTHREAD, str);
	fprintf(dbg_fp, "[tid %d] %s\n", MYTHREAD, str);
	fflush(dbg_fp);
	upc_global_exit(4);
}

void taskq_all_run(taskq_t *taskq) {
	DEBUG3( fprintf(dbg_fp, "[taskq_all_run] strt\n"); )
  do {
    while ( taskq_execute(taskq) != _FAIL ) { ; }   // executes all the tasks in the local task queue
    if ( taskq_steal(taskq) ) continue;    // tries to steal tasks from random victim threads
  } while ( !taskq_all_isEmpty(taskq) );    // checks the global termination condition
	DEBUG3( fprintf(dbg_fp, "[taskq_all_run] done\n"); )
}

void taskq_local_reset(taskq_t *taskq)
{
	rtaskq_t *rtaskq = taskq->rtaskq_lp;

	DEBUG3( fprintf(dbg_fp, "[taskq_local_reset] strt\n"); )

	upc_lock(rtaskq->lock);
	rtaskq->pcnt = 0;
	rtaskq->ptop = 0;
	rtaskq->scnt = 0;
	rtaskq->pbot = 0;
	rtaskq->tail = 0;
	rtaskq->prev_ptop = 0;
	upc_unlock(rtaskq->lock);
}

void taskq_set_steal_trial(taskq_t *taskq, int preset_steal_trial, int total_steal_trial) {

	if (preset_steal_trial > 0) taskq->preset_steal_trial = preset_steal_trial;
	if (total_steal_trial > 0) taskq->total_steal_trial = total_steal_trial;

	DEBUG1( 
		fprintf(dbg_fp, "[taskq_set_steal_trial] preset_steal_trial=%d\n", taskq->preset_steal_trial); 
		fprintf(dbg_fp, "                        total_steal_trial=%d\n", taskq->total_steal_trial); 
	)
}

void taskq_set_independent_task(taskq_t *taskq, void *func_ptr) {
	func_type_t *func_type_p;
	for (int i=0; i<taskq->numTaskFuncs; i++) {
		func_type_p = &taskq->task_func_list[i];
		if ((uint64_t)func_type_p->funcP == (uint64_t)func_ptr) {
			func_type_p->no_output_dep = 1;
			break;
		}
	}
}

void taskq_set_runtime_cutoff(taskq_t *taskq, int rc_coeff) {
	taskq->runtime_cutoff = rc_coeff;
}

void taskq_set_steal_size(taskq_t *taskq, int steal_size) {
	taskq->steal_size = steal_size;
	DEBUG1( fprintf(dbg_fp, "[taskq_set_steal_size] taskq->steal_size=%d\n", taskq->steal_size); )
}

void taskq_set_chunk_size(taskq_t *taskq, int chunk_size) {
	taskq->chunk_size = chunk_size;
	DEBUG1( fprintf(dbg_fp, "[taskq_set_chunk_size] taskq->chunk_size=%d\n", taskq->chunk_size); )
}

void taskq_set_threshold(taskq_t *taskq, int threshold) { 
	taskq->threshold = threshold; 
	if (threshold<2) {
		if (MYTHREAD==0) printf("[taskq_set_threshold] too small threshold(=%d)\n", threshold); 
		upc_global_exit(1);
	}

	DEBUG1( fprintf(dbg_fp, "[taskq_set_threshold] taskq->threshold=%d\n", taskq->threshold); )
	return; 
}

int get_funcID(taskq_t *taskq, void *func_ptr)
{
	func_type_t *func_type_p;
	for (int i=0; i<taskq->numTaskFuncs; i++) {
		func_type_p = &taskq->task_func_list[i];
		if ((uint64_t)func_type_p->funcP == (uint64_t)func_ptr) { return i; }
	}
	printf("[get_funcID] tid%d ERROR func_ptr=%p\n", MYTHREAD, func_ptr);
	upc_global_exit(1);
	return 0;
}

void taskq_set_cores_per_node(taskq_t *taskq, int ncores)
{
	int i;

	taskq->cores_per_node = ncores;
	if ( (THREADS%taskq->cores_per_node) != 0 ) {
		if (MYTHREAD==0) {
			printf("[taskq_set_cores_per_node] total UPC threads (THREADS=%d) ", THREADS);
			printf("is not a multiple of cores_per_node(%d)\n", taskq->cores_per_node);
		}
		upc_global_exit(1);
	}
	if ( (MYTHREAD%taskq->cores_per_node) == 0 ) i_am_a_local_root = 1;

	num_local_threads = 0;
	local_thread_list = (int *) malloc(taskq->cores_per_node * sizeof(int));

	DEBUG1( 
		fprintf(dbg_fp, "num_local_threads=%d\n", num_local_threads); 
		fprintf(dbg_fp, "local_thread_list = [");
		for (i=0; i<num_local_threads; i++) {
			if (i!=0) fprintf(dbg_fp, ", ");
			fprintf(dbg_fp, "%d", local_thread_list[i]);
		}
		fprintf(dbg_fp, "]\n");
	)
}

void taskq_set_first_steal_from_root(taskq_t *taskq) {
  taskq->first_steal_from_root = 1;
}

/* -----------------------------------------------------------
 * chunk_mode: FIXED_CHUNK=1, HALF_CHUNK=2, ADAPTIVE_CHUNK=3
 * ----------------------------------------------------------- */
void taskq_set_chunk_mode(taskq_t *taskq, int chunk_mode) {
	taskq->chunk_mode = chunk_mode;
}

/* -----------------------------------------------------------
 * taskq_set_steal_mode(taskq, FULL_RAND_STEAL, 1, THREADS);
 * taskq_set_steal_mode(taskq, SAME_NODE_STEAL, 1, nCores);
 * taskq_set_steal_mode(taskq, HIER_WORK_STEAL, 2, nCores, THREADS);
 * ----------------------------------------------------------- */
void taskq_set_steal_mode(taskq_t *taskq, int steal_mode, int num_hierarchy, ...) {
	int i;
	rtaskq_t *rtaskq_lp = taskq->rtaskq_lp;
	va_list ap;

	DEBUG1( fprintf(dbg_fp, "[taskq_set_steal_mode] steal_mode=%d\n", steal_mode); )

	if (THREADS<2) return;

 /* --------------- variable arguments starts ---------------- */
  va_start(ap, num_hierarchy);

	taskq->steal_mode = steal_mode;
	taskq->num_hierarchy = num_hierarchy;
	taskq->nprocs_per_level = (int *) malloc(num_hierarchy * sizeof(int));

  for (i=0; i<num_hierarchy; i++) {
		taskq->nprocs_per_level[i] = (int)va_arg(ap, int);
		if (taskq->nprocs_per_level[i] < taskq->nprocs_per_level[i-1]) {
			taskq_error_exit("[taskq_set_steal_mode] ERROR: level_nprocs should be an increaing order\n");
		}
		if (taskq->nprocs_per_level[i] > THREADS) {
			taskq_error_exit("[taskq_set_steal_mode] ERROR: level_nprocs cannot be larger than THREADS\n");
		}
  }

  va_end(ap);

  /* --------------- variable arguments done ---------------- */

	if (steal_mode==SAME_NODE_STEAL) {
		taskq->preset_steal_trial = taskq->total_steal_trial;		// update preset_steal_trial
	}

	taskq->victim_list = gen_victim_list(taskq);
	DEBUG1(
		fprintf(dbg_fp, "             ");
		for (i=0; i<taskq->total_steal_trial; i++) { fprintf(dbg_fp, "%3d ", i); }
		fprintf(dbg_fp, "\n");
		fprintf(dbg_fp, "victim_list=[");
		for (i=0; i<taskq->total_steal_trial; i++) { fprintf(dbg_fp, "%3d ", taskq->victim_list[i]); }
		fprintf(dbg_fp, "]\n");
	)
}

int *init_random_victim_list(taskq_t *taskq)
{
	int i;
	int *victim_list;

	DEBUG1( 
		fprintf(dbg_fp, "[init_random_victim_list] total_steal_trial=%d\n", taskq->total_steal_trial); 
	)

	if (taskq->victim_list) free(taskq->victim_list);
	victim_list = (int *)malloc( (taskq->total_steal_trial+1)*sizeof(int) );
	for (i=0; i<taskq->total_steal_trial; i++) { victim_list[i] = -1;	}

	return victim_list;
}

int *gen_victim_list(taskq_t *taskq)
{
	int i, j, k;
	double scale;
  int vlist_dbg_cnt;
  int vcnt, bottom, offset, candidate;
	double min=99999.9, max=0.0, avg=0.0;
	int duplicate;
	int s_window;		// start of the window to check any duplicated victim candidate 
	int cur_level;
	int steal_mode = taskq->steal_mode;
	int num_hierarchy = taskq->num_hierarchy;
	int *nprocs_per_level = taskq->nprocs_per_level;
	int cumulative_levels = nprocs_per_level[0];
	int *victim_list;

	DEBUG3( 
		fprintf(dbg_fp,"[gen_victim_list] preset_steal_trial=%d ", taskq->preset_steal_trial);
		fprintf(dbg_fp,"total_steal_trial=%d, num_hierarchy=%d\n",taskq->total_steal_trial,num_hierarchy); 
	)

	victim_list = init_random_victim_list(taskq);

	j = MYTHREAD;	// need this initialization for RAND_NEXT
	vcnt = 0;
	cur_level = 0;
	vlist_dbg_cnt = 0;

	if ( (steal_mode==HIER_WORK_STEAL) || (steal_mode==SAME_NODE_STEAL) ) {
		while (vcnt < taskq->preset_steal_trial) {
			duplicate = 0;

			DEBUG1( fprintf(dbg_fp, "\nvcnt=%d --------------------------\n", vcnt); )

		  if ( (vcnt >= cumulative_levels && cur_level < (num_hierarchy-1)) ||
					 (vcnt==1 && (nprocs_per_level[cur_level]<=2)) )
			{
				cur_level++;
				cumulative_levels += nprocs_per_level[cur_level];
			}

			bottom = (MYTHREAD / nprocs_per_level[cur_level]) * nprocs_per_level[cur_level];

#if 0  /* random steal */
			scale = rand()/(float)RAND_MAX;   /* scale = [0:1] */
			j = (int)(scale * nprocs_per_level[cur_level]) + bottom; 
			j = (j%THREADS);

			if (cur_level < num_hierarchy)
				s_window = vcnt - (int)(nprocs_per_level[cur_level]/2);
			else
				s_window = vcnt - (int)(nprocs_per_level[num_hierarchy-1]/2);

			if (s_window < 0) s_window = 0;

			for (k=s_window; k<vcnt; k++) { 	
				if (j == victim_list[k]) { duplicate = 1; break; } 
			}

			if (j!=MYTHREAD && !duplicate) {
				victim_list[vcnt++] = j;
			}

			DEBUG1( fprintf(dbg_fp, "bot=%d scale=%f s_wind=%d cur_level=%d j=%d\n",
							bottom, scale, s_window, cur_level, j); )

			DEBUG1(
				if (vlist_dbg_cnt++ > taskq->preset_steal_trial*20) { 
					fprintf(dbg_fp, "ERROR: vlist_dbg_cnt=%d\n", vlist_dbg_cnt);
					upc_global_exit(1);
				}
			)
#else
			j = ((j + 1)%nprocs_per_level[cur_level] + bottom);
			j = (j%THREADS);
			if (j != MYTHREAD) victim_list[vcnt++] = j;
#endif
		}
	}

	DEBUG0(
		fprintf(dbg_fp, "             ");
		for (i=0; i<taskq->total_steal_trial; i++) { fprintf(dbg_fp, "%3d ", i); }
		fprintf(dbg_fp, "\n");
		fprintf(dbg_fp, "victim_list=[");
		for (i=0; i<taskq->total_steal_trial; i++) { fprintf(dbg_fp, "%3d ", victim_list[i]); }
		fprintf(dbg_fp, "]\n");
	)

	return victim_list;
}

taskq_handle taskq_all_alloc(int numTaskFuncs, ...)
{
	int i, j, k, ii;
	float scale;
	char fname[20];
	va_list ap;
	taskq_handle taskq;
	int rq_size = READYQ_SZ;
	int wq_buckets = WAITQ_BUCKETS;
	int wq_entries = WAITQ_ENTRIES;
	int msgbuf_sz = MSGBUF_SZ;
	int max_iSize=0, max_oSize=0;

	/* dbg_fp FILE pointer should have been opened before used */
	sprintf(fname, "p%d-taskq.dbg", MYTHREAD); 
	dbg_fp = fopen(fname, "w");

#ifdef ADAPTIVE_CHUNK_SIZE 
#ifndef __BERKELEY_UPC__
	if(MYTHREAD==0) printf("ADAPTIVE_CHUNK_SIZE is not supported on non-bupc environment\n");
	upc_global_exit(1);
#endif
#endif

	/* ================= taskq data structure allocation ================= */
	if ( (taskq = malloc(sizeof(taskq_t))) == NULL )  {
		printf("UPC thread %d: failed to allocate taskq\n", MYTHREAD); 
		upc_global_exit(1); 
	}

	DEBUG1( 
		fprintf(dbg_fp, "---------------------------------------------------------\n"); 
		fprintf(dbg_fp, "[taskq_all_alloc] numTaskFuncs=%d\n", numTaskFuncs);
		fprintf(dbg_fp, "[taskq_all_alloc] rq_size=%d wq_buckets=%d ", rq_size, wq_buckets);
		fprintf(dbg_fp, "wq_entries=%d msgbuf_sz=%d\n", wq_entries, msgbuf_sz); 
	)

	DEBUG3(
		if (THREADS > 100) { 
			if (MYTHREAD==0) printf("DO NOT RUN with DEBUG3 mode using large number of threads\n");
			fprintf(dbg_fp, "DO NOT RUN with DEBUG3 mode using large number of threads\n");
			upc_global_exit(1);
		}
	)

	/* --------------- Task Function ID Map --------------- */
	// allocate a list of task function types
	taskq->numTaskFuncs = numTaskFuncs;
	taskq->task_func_list = (func_type_t *)malloc(numTaskFuncs*sizeof(func_type_t));

	/* --------------- variable arguments starts ---------------- */
	va_start(ap, numTaskFuncs);

	for (ii=0; ii<numTaskFuncs; ii++) {
		func_type_t *func_type_p = &taskq->task_func_list[ii];
		uint64_t task_func_addr = (uint64_t)va_arg(ap, uint64_t);	// read one task func address
		func_type_p->funcP = (void (*)(void *, void*))task_func_addr;	// task func. addr. in this thread
		func_type_p->no_output_dep = 0;		// optional: the task function has no output dependency
		func_type_p->iSize = (int)va_arg(ap, int);		// input size of the task function
		func_type_p->oSize = (int)va_arg(ap, int);		// output size of the task function
		if (func_type_p->iSize > max_iSize) max_iSize = func_type_p->iSize;
		if (func_type_p->oSize > max_oSize) max_oSize = func_type_p->oSize;
		DEBUG0( fprintf(dbg_fp, "[taskq_all_alloc] funcID=%d: func=%lu iSize=%d oSize=%d\n", 
						ii, task_func_addr, func_type_p->iSize, func_type_p->oSize);
		)
	}
	va_end(ap);
	/* --------------- variable arguments done ---------------- */

	int	msg_sz  = sizeof(task_msg_header_t) + max_oSize;
	int task_sz = sizeof(task_header_t) + max_iSize;

	DEBUG0( 
		fprintf(dbg_fp, "[taskq_all_alloc] task_sz=%d msg_sz=%d\n", task_sz, msg_sz); 
		fprintf(dbg_fp, "[taskq_all_alloc] max_iSize=%d max_oSize=%d\n", max_iSize, max_oSize); 
	)

	/* ----------------- Ready Task Queue ----------------- */
	/* globally shared allocation */
	shared rtaskq_t *rtaskq_sp = (shared rtaskq_t *) upc_all_alloc(THREADS, sizeof(rtaskq_t));
	rtaskq_t *rtaskq_lp = (rtaskq_t *) &rtaskq_sp[MYTHREAD];
	shared char *rtaskq_gbuf = (shared char *)upc_all_alloc(THREADS, rq_size*task_sz);

	/* rtaskq related initialization */
	taskq->task_sz = task_sz;
	taskq->max_iSize = max_iSize;
	taskq->msg_sz = msg_sz;
	taskq->taskID = 0;
	taskq->syncID = 0;		
	taskq->steal_size = DEFAULT_TASKQ_STEAL_SIZE;
	taskq->chunk_size = DEFAULT_TASKQ_CHUNK_SIZE;
	taskq->threshold = DEFAULT_TASKQ_THRESHOLD;

	taskq->total_steal_trial = 4*ilog2(THREADS);
	int random_steal_trial = ilog2(THREADS);
	if (random_steal_trial > 0) 
		taskq->preset_steal_trial = taskq->total_steal_trial - ilog2(THREADS);
	else
		taskq->preset_steal_trial = taskq->total_steal_trial - 1;

	/* task_get_buf: instead of task_sz, allocate task_head_t size (not sure thou) */
	taskq->task_get_buf = (task_header_t *) malloc(TASK_STACK_SZ * task_sz);
	taskq->task_put_buf = (task_header_t *) malloc(task_sz);
	taskq->task_out_buf = (char *) malloc(TASK_STACK_SZ * max_oSize);
	taskq->task_msg_buf = (task_msg_header_t *) malloc(msg_sz);

	DEBUG3( fprintf(dbg_fp, "taskq->task_get_buf: TASK_STACK_SZ=%d, task_sz=%d, saddr=%p sizeof(taskq->task_get_buf)=%llu\n", TASK_STACK_SZ, task_sz, taskq->task_get_buf, sizeof(taskq->task_get_buf)); ) 

	taskq->rtaskq_sp = rtaskq_sp;
	taskq->rtaskq_lp = rtaskq_lp;

	rtaskq_lp->size = rq_size;
	rtaskq_lp->lock = upc_global_lock_alloc();		/* lock should be allocated before taskq_local_reset */

	taskq_local_reset(taskq);
	// rtaskq_lp->pcnt = 0;				/* reset in the taskq_local_reset */
	// rtaskq_lp->ptop = 0;				/* reset in the taskq_local_reset */
	// rtaskq_lp->scnt = 0;				/* reset in the taskq_local_reset */
	// rtaskq_lp->pbot = 0;				/* reset in the taskq_local_reset */
	// rtaskq_lp->tail = 0;				/* reset in the taskq_local_reset */
	// rtaskq_lp->prev_ptop = 0; 	/* reset in the taskq_local_reset */
	rtaskq_lp->nAcquire = 0;
	rtaskq_lp->nRelease = 0;
	rtaskq_lp->failed_victim_probe = 0;
	rtaskq_lp->failed_steal_attempt = 0;
	rtaskq_lp->nSteal = 0;
	rtaskq_lp->nTaskSteal = 0;
	rtaskq_lp->nWakeUp = 0;
	rtaskq_lp->nFail = 0;
	rtaskq_lp->nPut = 0;
	rtaskq_lp->nGet = 0;

	rtaskq_lp->nProduced = upc_all_alloc(THREADS, sizeof(int));
	rtaskq_lp->nConsumed = upc_all_alloc(THREADS, sizeof(int));
	rtaskq_lp->gbuf = (shared char *) &rtaskq_gbuf[MYTHREAD];
	rtaskq_lp->gbuf_lp = (char *) rtaskq_lp->gbuf;

	/* --------------- Task Message Buffer --------------- */
	task_msgbuf_sp msgbuf_sp = (task_msgbuf_sp) upc_all_alloc(THREADS, sizeof(task_msgbuf_t));
	task_msgbuf_lp msgbuf_lp = (task_msgbuf_lp) &msgbuf_sp[MYTHREAD];
	msgbuf_lp->size = msgbuf_sz;
/*
	msgbuf_lp->msg_cnt = (shared int *) upc_alloc(sizeof(int));
	*(msgbuf_lp->msg_cnt) = 0;
*/
	msgbuf_lp->lock = upc_global_lock_alloc();
	shared char *msgbuf_gbuf = (shared char *) upc_all_alloc(THREADS, msgbuf_sz*msg_sz);
	msgbuf_lp->sbuf = (shared char *) &msgbuf_gbuf[MYTHREAD];
	msgbuf_lp->sbuf_lp = (char *) msgbuf_lp->sbuf;

	/* msgbuf related initialization */
	taskq->msgbuf_sp = msgbuf_sp;
	taskq->msgbuf_lp = msgbuf_lp;

	/* temporary thing */
/*
	msgbuf_cnt = (shared int *) upc_all_alloc(THREADS, sizeof(int));
*/
	msgbuf_cnt[MYTHREAD] = 0;

	/* --------------- Wait Task Queue Map --------------- */
	taskq->wait_task_map = (upc_map_t *) upc_map_alloc("wait_task_map", wq_buckets, wq_entries, task_sz);

	/* --------------- Sync Node Map --------------- */
	taskq->sync_node_map = (upc_map_t *) upc_map_alloc("sync_node_map", wq_buckets, wq_entries, sizeof(sync_node_t));

	/* --------------- initialize rand --------------- */
	//srand(MYTHREAD);
	srand( MYTHREAD*time(NULL) );

	// ------------------------------------------------------------------------
	// initialize global variables
	// ------------------------------------------------------------------------
	i_am_a_local_root = 0;
	taskq->cores_per_node = 0;		// 0 indicates that cores_per_node is not specified

	// ------------------------------------------------------------------------
	// initialize steal related variables
	// ------------------------------------------------------------------------
	taskq->steal_mode = FULL_RAND_STEAL;
	taskq->num_hierarchy = 1;
	taskq->victim_list = init_random_victim_list(taskq);
	taskq->nprocs_per_level = NULL;

  taskq->first_steal_from_root = 0;

	// ------------------------------------------------------------------------
	// initialize taskq variable for task stack 
	// ------------------------------------------------------------------------
	taskq->task_stack_top = 0;
	taskq->task_stack_max = 0;

	// ------------------------------------------------------------------------
	// record victim process IDs
	// ------------------------------------------------------------------------
	victim_count = (int *)malloc(sizeof(int)*THREADS);
	for (i=0; i<THREADS; i++) victim_count[i] = 0;

	// ------------------------------------------------------------------------
	// intialize variables for taskq_isEmpty()
	// ------------------------------------------------------------------------
	cb_lock = upc_all_lock_alloc();		// init a single lock (should be upc_all_lock_alloc)

	if (MYTHREAD==0) {
		upc_lock(cb_lock);
		cb_count = 0;
		cb_cancel = 0;
		cb_done = 0;
		upc_unlock(cb_lock);
	}

	// ------------------------------------------------------------------------
	// set castable_threads if BUPC is available
	// ------------------------------------------------------------------------
	num_local_threads = 0;
	local_thread_list = NULL;
#ifdef __BERKELEY_UPC__
	for (i=1; i<THREADS; i++) {
		j = (MYTHREAD + i) % THREADS;
		if (bupc_castable( &rtaskq_sp[j] )) { 
			num_local_threads++;
		}
	}
	local_thread_list = (int *) malloc(num_local_threads * sizeof(int));
	num_local_threads = 0;
	for (i=1; i<THREADS; i++) {
		j = (MYTHREAD + i) % THREADS;
		if (bupc_castable( &rtaskq_sp[j] )) { 
			local_thread_list[num_local_threads++] = j;
		}
	}
#endif

#ifdef HOPPER
	if (MYTHREAD==0) printf("[HOPPER mode]\n");
	num_local_threads = 24;
	int bottom = (MYTHREAD / num_local_threads) * num_local_threads;
	for (i=0; i<num_local_threads; i++) {
		local_thread_list[i] = bottom + i;
	}
#endif

	/* this barrier ensures that memory is all allocated before the application starts */
	upc_barrier;

	DEBUG1( 
		fprintf(dbg_fp, "[taskq_all_alloc] rtaskq: gbuf=%lu gbuf_lp=[%p : %p]\n", rtaskq_lp->gbuf, 
						rtaskq_lp->gbuf_lp, rtaskq_lp->gbuf_lp + rq_size*task_sz);
		fprintf(dbg_fp, "  msgbuf: thread local sbuf=[%lu : %lu]\n", msgbuf_lp->sbuf, 
						msgbuf_lp->sbuf + msgbuf_sz*msg_sz);
		fprintf(dbg_fp, "  initial taskq->total_steal_trial=%d\n", taskq->total_steal_trial);
		fprintf(dbg_fp, "  initial taskq->preset_steal_trial=%d\n", taskq->preset_steal_trial);
		fprintf(dbg_fp, "---------------------------------------------------------\n"); 
	)

#ifdef ENABLE_TIMER
	timer_clear(PUT_TIME);
#endif

	return taskq;
}


// set the cb_cancel variable to wake-up threads that are spin-waiting on the taskq_isEmpty()
void taskq_set_notAllEmpty()
{
#ifdef __BERKELEY_UPC__
	bupc_waitsync(cb_handle);
	cb_handle = bupc_memput_async((shared void *)&cb_cancel, (const void *)&local_cb_cancel, sizeof(int));
#else
	cb_cancel = 1;
#endif
}

// returns 1 if the local ready queue is empty (even if the wait queue is not empty) 
int taskq_isEmpty(taskq_t *taskq) {
	int result = EMPTY;
	rtaskq_t *rtaskq_lp = taskq->rtaskq_lp;

	DEBUG1( fprintf(dbg_fp, "[taskq_isEmpty] strt --------------------------------\n"); )

	if (rtaskq_lp->pcnt > 0) { 
		DEBUG1( fprintf(dbg_fp, " returns NOT EMPTY! (rtaskq_lp->pcnt=%d)\n", rtaskq_lp->pcnt); )
		result = NOT_EMPTY; 
	}
	else if (rtaskq_lp->scnt > 0 && taskq_acquire(taskq) > 0 ) { 
		// this check is practically useless, but for correctness.
		DEBUG1( fprintf(dbg_fp, " returns NOT EMPTY! (rtaskq_lp->scnt=%d)\n", rtaskq_lp->scnt); )
		result = NOT_EMPTY;
	}
	else {
		UPC_POLL_NETWORK;
		if (msgbuf_cnt[MYTHREAD] > 0) {
			DEBUG3(fprintf(dbg_fp, "[taskq_isEmpty] msgbuf_cnt[%d]=%d\n", MYTHREAD, msgbuf_cnt[MYTHREAD]);)
			if (taskq_process_msgbuf(taskq) > 0) {
				if (taskq->rtaskq_lp->pcnt > 0) { 
					result = NOT_EMPTY; 
				}
			}
		}
	}

	DEBUG1( fprintf(dbg_fp, "[taskq_isEmpty] done: returns %d -------------------------\n", result); )
	return result;
}

int taskq_all_isEmpty(taskq_t *taskq) {
	int flag;
	int l_cancel, l_done;
	char head[20] = "taskq_all_isEmpty";
	int dbg_count=0;
	task_header_t *task_p;
	task_header_t *wait_task_p;
	sync_node_t *sync_node;
	int i;

	DEBUG1( fprintf(dbg_fp, "[%s] strt --------------------------------\n", head); )

	// STEP 1: check if the taskq is locally empty
	if ( taskq_isEmpty(taskq) == NOT_EMPTY ) return NOT_EMPTY;

	DEBUG3( fprintf(dbg_fp, "[%s] 1: # wait_task=%d\n", head, taskq->wait_task_map->entry_cnt); )

	// STEP 2: spin-wait until the wait taskq is empty
	while ( taskq->wait_task_map->entry_cnt > 0) {
		UPC_POLL_NETWORK;
		if (++dbg_count > 100000) {
			task_p = upc_map_get_one(taskq->wait_task_map);
			fprintf(dbg_fp, "[%s] ERROR: there is no progress, it is stuck!\n", head);
			fprintf(dbg_fp, "[%s] ERROR: # wait_task=%d\n", head, taskq->wait_task_map->entry_cnt);
			fprintf(dbg_fp, "  a wait_task (taskID=%lu syncID=%lu)\n", task_p->taskID, task_p->syncID);
			sync_node = upc_map_get(taskq->sync_node_map, task_p->syncID);
			fprintf(dbg_fp, "  sync_node=%p\n", sync_node);
			fprintf(dbg_fp, "  sync_node->ref_count=%d\n", sync_node->ref_count);
			for (i=0; i<sync_node->ref_count; i++) {
				wait_task_p = sync_node->wait_task;
				fprintf(dbg_fp, "[taskq_all_isEmpty] wait_task_p->taskID=%lu wait_task_p->syncID=%lu\n", 
								wait_task_p->taskID, wait_task_p->syncID);
				wait_task_p = wait_task_p->next;
			}					
			upc_global_exit(1);
		}
		if (msgbuf_cnt[MYTHREAD] > 0) {
			if (taskq_process_msgbuf(taskq) > 0) {
				if (taskq->rtaskq_lp->pcnt > 0) return NOT_EMPTY;
			}
		}
	}

	DEBUG2( fprintf(dbg_fp, "[%s] 2: cb_count=%d\n", head, cb_count); )

	flag = cbarrier_inc(taskq);
	DEBUG3( fprintf(dbg_fp, "[%s] 3: cb_count=%d\n", head, cb_count); )

	while (flag != TERM) {
		UPC_POLL_NETWORK;
		if (detectWork(taskq) && cbarrier_dec(taskq) != TERM) {
			flag = NO_TERM;
			taskq->rtaskq_lp->nWakeUp++;
			break;
		}
		if (cb_count == THREADS) flag = TERM;
		else flag = NO_TERM;
	}			

	DEBUG3( fprintf(dbg_fp, "[%s] 4: cb_count=%d flag=%d\n", head, cb_count, flag); )

	upc_lock(cb_lock);
	if (cb_count == 0) {				// the last leaving thread cleans the mess
		if (cb_done) cb_done = 0;;
		if (cb_cancel) cb_cancel = 0;;
	}
	upc_unlock(cb_lock);

	if (flag == TERM) {
		taskq_local_reset(taskq);
		upc_barrier;
		if (MYTHREAD==0) cb_count = 0;
	}

	DEBUG3( fprintf(dbg_fp, "[%s] done returns %d -------------------------\n", head, flag); )
	return flag;
}

void taskq_wait(taskq_handle taskq)
{
	int dbg_count=0;
	sync_node_t sync_node;
	sync_node_t *sync_node_p;
	task_header_t *parent_task = (task_header_t *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top-1)*taskq->task_sz);
	uint64_t cur_syncID = parent_task->child_syncID;

	DEBUG3( fprintf(dbg_fp, "[taskq_wait] (parent taskID=%lu, parent syncID=%lu) strt -------------\n", 
					parent_task->taskID, parent_task->syncID); )

	if (parent_task->sync_tcnt == 0) {
		DEBUG2( fprintf(dbg_fp, "[taskq_wait] returns immediately (sync_tcnt=%d)\n", parent_task->sync_tcnt); )
		return;
	}

	sync_node.ref_count = parent_task->sync_tcnt;
	sync_node.wait_task = NULL;	// cannot have a task in the wait queue waiting this sync_node 
	sync_node_p = upc_map_put(taskq->sync_node_map, cur_syncID, &sync_node);

	DEBUG3( fprintf(dbg_fp, "[taskq_wait] insert sync_node syncID=%lu\n", cur_syncID); )

	if ( (++taskq->syncID) < 0 ) {
		printf("[taskq_wait] tid%d: syncID overflow! taskID=%lu\n", MYTHREAD, taskq->syncID);
		upc_global_exit(1);
	}

	parent_task->prev_child_syncID = parent_task->child_syncID;		// update prev_child_syncID
	parent_task->child_syncID = taskq->syncID;							// update child_syncID
	parent_task->sync_tcnt = 0;
	parent_task->sync_phase = 0;

	// this is unsafe code: sync_node_p points to the map entry where sync_node resides.
	// Inside taskq_execute and taskq_isEmpty call upc_map_remove functions when ref_count becomes zero
	// Therefore, sync_node_p could get easily corrupted if there is any upc_map functions interfering.  
	// solution: should taskq_wait remove sync_node from sync_node_map?
/*
	while (sync_node_p->ref_count > 0) {
		do {
			if ( sync_node_p->ref_count == 0 ) break;
			while ( taskq_execute(taskq) != _FAIL ) { 
				if ( sync_node_p->ref_count == 0 ) break; // need this otherwise stack overflow!
			}
			if ( sync_node_p->ref_count == 0 ) break;		// need this to prevent calling taskq_steal
			if (taskq_steal(taskq) > 0) continue;
		} while ( !taskq_isEmpty(taskq) );	// remove this while-loop replace isEmpty check with sthelse
	}
*/
	while ( upc_map_get(taskq->sync_node_map, cur_syncID) != NULL ) {
		do {
			if ( upc_map_get(taskq->sync_node_map, cur_syncID) == NULL ) break;
			while ( taskq_execute(taskq) != _FAIL ) { 
				// need this otherwise stack overflow!
				DEBUG3( fprintf(dbg_fp, "[taskq_wait] 1: parent_task->taskID=%lu cur_syncID=%lu\n", 
								parent_task->taskID, cur_syncID); )
				if ( upc_map_get(taskq->sync_node_map, cur_syncID) == NULL ) {
					DEBUG3( fprintf(dbg_fp, "[taskq_wait] 2: cur_syncID=%lu complete\n", cur_syncID); )
					break;
				}
				DEBUG3( fprintf(dbg_fp, "[taskq_wait] 3: cur_syncID=%lu\n", cur_syncID); )
			}
			DEBUG3( fprintf(dbg_fp, "[taskq_wait] 4: cur_syncID=%lu\n", cur_syncID); )
			// need this to prevent calling taskq_steal
			if ( upc_map_get(taskq->sync_node_map, cur_syncID) == NULL ) {
				DEBUG3( fprintf(dbg_fp, "[taskq_wait] 5: cur_syncID=%lu complete\n", cur_syncID); )
				break;
			}
			if ( taskq_steal(taskq) > 0 ) continue;
		} while ( !taskq_isEmpty(taskq) );	// remove this while-loop replace isEmpty check with sthelse
	}

	DEBUG3( fprintf(dbg_fp, "[taskq_wait] cur_syncID=%lu done\n", cur_syncID); )
	return;
}

void taskq_fence(taskq_t *taskq)
{
	task_header_t *task_p;
	sync_node_t sync_node;
	//task_header_t *parent_task = &taskq->task_get_buf[taskq->task_stack_top-1];
	task_header_t *parent_task = (task_header_t *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top-1)*taskq->task_sz);

	DEBUG3( 
		fprintf(dbg_fp, "[taskq_fence] strt (sync_tcnt=%d) ----------------\n", parent_task->sync_tcnt); 
	)

	if (parent_task->sync_tcnt == 0) {
		DEBUG2( fprintf(dbg_fp, "[taskq_fence] returns immediately (sync_tcnt=%d)\n", parent_task->sync_tcnt); )
		return;
	}

	sync_node.ref_count = parent_task->sync_tcnt;
	sync_node.wait_task = NULL;		// it will be added by the following taskq_put functions
	upc_map_put(taskq->sync_node_map, parent_task->child_syncID, &sync_node);
	DEBUG3( fprintf(dbg_fp, "[taskq_fence] put sync_node syncID=%lu\n", parent_task->child_syncID); )

	// update syncID and sync_phase, and reset sync_tcnt
	if ( (++taskq->syncID) < 0 ) {
		printf("[taskq_fence] tid%d: syncID overflow! taskID=%lu\n", MYTHREAD, taskq->syncID);
		upc_global_exit(1);
	}

	parent_task->prev_child_syncID = parent_task->child_syncID;		// update prev_child_syncID
	parent_task->child_syncID = taskq->syncID;										// update child_syncID
	parent_task->sync_phase++;
	parent_task->sync_tcnt = 0; 

	DEBUG3( fprintf(dbg_fp, "[taskq_fence] done -------------------------------------\n"); )
}

// we need taskq_put_continuation because otherwise we cannot tell if it is the last task, 
// which contains the parent function's tail (i.e., returns the parent function).
int taskq_put_continuation(taskq_handle taskq, void *funcPtr, ...)
{
	va_list ap;
	int ret;
	void *dst;
	void (*func)(void *, void *) = NULL;
	func_type_t *func_type;
	task_header_t *task_p = taskq->task_put_buf;
	task_header_t *parent_task = (task_header_t *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top-1)*taskq->task_sz);

	DEBUG3( fprintf(dbg_fp, "[taskq_put_cont] strt (parent_task->taskID=%lu) ------\n", parent_task->taskID); )

#ifdef ANALYSIS
	rtaskq_snapshot(taskq->rtaskq_lp);
#endif

	parent_task->hasCont = 1;		// mark that the parent task will be replaced with this cont. task

	// ------------ variable arguments start ------------ //
	va_start(ap, funcPtr);	// initialize ap and informs the last named parameter, funcPtr.

	if ( (++taskq->taskID) < 0) {
		printf("[taskq_put] tid%d: taskID overflow! taskID=%lu\n", MYTHREAD, taskq->taskID);
		upc_global_exit(1);
	}
	task_p->taskID = taskq->taskID;
	task_p->syncID = parent_task->syncID;
	task_p->child_syncID = parent_task->child_syncID;
	task_p->prev_child_syncID = parent_task->child_syncID;
	task_p->funcID = get_funcID(taskq, funcPtr);
	func_type = &taskq->task_func_list[task_p->funcID];
	task_p->sync_tcnt = 0;
	task_p->sync_phase = 0;
	task_p->owner = parent_task->owner;
	//task_p->inlined = parent_task->inlined;		// should not have meaning for a continuation task
	task_p->func = funcPtr;
	task_p->hasCont = 0;
	task_p->out_size = func_type->oSize;
	task_p->out_offs = parent_task->out_offs;	
	task_p->next = NULL;	

	DEBUG3( fprintf(dbg_fp, "[taskq_put_cont] funcID=%d iSize=%d oSize=%d\n", task_p->funcID, func_type->iSize, func_type->oSize); )

	// if the task function has input data, read input address
	if ( func_type->iSize > 0) { task_p->iptr = (void *)va_arg(ap, void *); }

	DEBUG3( fprintf(dbg_fp, "[taskq_put_cont] parent_task->optr=%p\n", parent_task->optr); )
	task_p->optr = parent_task->optr;	// parent task's optr

	va_end(ap);
	// ------------ variable arguments ends ------------ //

#ifdef RUNTIME_CUTOFF
	// if parent task's sync_phase==0, then taskq_put_continuation is the same as taskq_put
  if ((parent_task->sync_phase==0) &&
      ((taskq->rtaskq_lp->scnt + taskq->rtaskq_lp->pcnt) > taskq->runtime_cutoff) ) {


		DEBUG3( fprintf(dbg_fp, "[taskq_put_cont] iptr=%p optr=%p\n", task_p->iptr, task_p->optr); )
		if ((++rcutoff_count%10)==0) { UPC_POLL_NETWORK; }
	
		DEBUG3( fprintf(dbg_fp, "[serialize] Inline start (taskID=%lu syncID=%lu)\n", 
						task_p->taskID, task_p->syncID); )
	
		taskq->task_stack_top++;
		//task_p->inlined = 1;
		if (taskq->task_stack_top >= TASK_STACK_SZ) {
			printf("[taskq_put_continuation] OVERFLOW ERROR: MYTHREAD=%d task_stack_top=%d\n", 
							MYTHREAD, taskq->task_stack_top);
			upc_global_exit(1);
		}
		dst = (void *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top-1)*taskq->task_sz);
	
		memcpy(dst, task_p, sizeof(task_header_t));		// copy only the header (no input data copied)
	
		DEBUG3( fprintf(dbg_fp, "[serialize] taskID=%lu syncID=%lu task_stack_top=%d\n", 
						task_p->taskID, task_p->syncID, taskq->task_stack_top); )
		//func = (void (*)(void *, void *))(task_p->func);
		memcpy(func, task_p->func, sizeof(func));
		func(task_p->iptr, task_p->optr);		// has no wait_task and if it is my task
			
		DEBUG3( fprintf(dbg_fp, "[serialize] Inline done (taskID=%lu syncID=%lu)\n", 
						task_p->taskID, task_p->syncID); )
		taskq->task_stack_top--;
		ret = 0;
	}
	else
		ret = taskq_insert(taskq, task_p, func_type);
#else
	ret = taskq_insert(taskq, task_p, func_type);
#endif
	DEBUG3( fprintf(dbg_fp, "[taskq_put_continuation] done (taskID=%lu parent=%lu syncID=%lu\n", 
					task_p->taskID, parent_task->taskID, task_p->syncID); )

	parent_task->sync_tcnt = 0;
	return ret;
}


// Let upc_map_put(wait_queue) return an address of the parent node 
// Then use this returned address of the parent node as the value of
// the key (child node) to be inserted at dep_graph_map <--- old story
int taskq_put(taskq_handle taskq, void *funcPtr, ...)
{
	int canary1 = 12345;
	int ret;
	va_list ap;
	void *dst;
	void (*func)(void *, void *) = NULL;
	func_type_t *func_type;
	task_header_t *task_p = taskq->task_put_buf;
	task_header_t *parent_task = NULL;

#ifdef ENABLE_TIMER
	timer_start(PUT_TIME);
#endif

	if (taskq->task_stack_top > 0) {
		parent_task = (task_header_t *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top-1)*taskq->task_sz);
	}

	DEBUG3( 
		if (parent_task==NULL)
			fprintf(dbg_fp, "[taskq_put] root task strt ------------------\n");
		else
			fprintf(dbg_fp, "[taskq_put] parent taskID=%lu strt ------------------\n", parent_task->taskID);
	)

	//debug_stack(taskq, "[taskq_put] 1 ");

#ifdef ANALYSIS
	rtaskq_snapshot(taskq->rtaskq_lp);
#endif

	// ------------ variable arguments start ------------ //
	va_start(ap, funcPtr);	// initialize ap and informs the last named parameter, funcPtr.

	task_p->funcID = get_funcID(taskq, funcPtr);
	func_type = &taskq->task_func_list[task_p->funcID];

	// if the task function has input, read input pointer
	if ( func_type->iSize > 0) {	task_p->iptr = (void *)va_arg(ap, void *); }

	// if the task function has output, read output address
	if ( func_type->oSize > 0) {	task_p->optr = (void *)va_arg(ap, void *); }

	va_end(ap);
	// ------------ variable arguments ends ------------ //

	if ( (++taskq->taskID) < 0) {
		printf("[taskq_put] tid%d: taskID overflow! taskID=%lu\n", MYTHREAD, taskq->taskID);
		upc_global_exit(1);
	}
	task_p->taskID = taskq->taskID;
	task_p->sync_tcnt = 0;
	task_p->sync_phase = 0;
	task_p->owner = MYTHREAD;
	//task_p->inlined = 0;
	task_p->func = funcPtr;
	task_p->hasCont = 0;
	task_p->out_size = func_type->oSize;
	task_p->out_offs = 0;
	task_p->next = NULL;	

	if (parent_task == NULL) {
		task_p->syncID = taskq->syncID;
		task_p->child_syncID = taskq->syncID;
		task_p->prev_child_syncID = taskq->syncID;
	}
	else {
		// assign a new syncID to the first task spawned under the parent task
		if (parent_task->syncID == parent_task->child_syncID) {
			if ( parent_task->child_syncID == taskq->syncID) {
				if ( (++taskq->syncID) < 0 ) {
					printf("[taskq_put] tid%d: syncID overflow! taskID=%lu\n", MYTHREAD, taskq->syncID);
					upc_global_exit(1);
				}
			}
			parent_task->prev_child_syncID = taskq->syncID;	// update the parent task's prev_child_syncID 
			parent_task->child_syncID = taskq->syncID;			// update the parent task's child_syncID 
		}
		task_p->syncID = parent_task->child_syncID;
		task_p->child_syncID = parent_task->child_syncID;
		task_p->prev_child_syncID = parent_task->child_syncID;
	}

	DEBUG3( fprintf(dbg_fp, "[taskq_put] taskID=%lu syncID=%lu iSize=%d oSize=%d iptr=%p optr=%p\n", 
	task_p->taskID, task_p->syncID, func_type->iSize, func_type->oSize, task_p->iptr, task_p->optr); )

#ifdef RUNTIME_CUTOFF
	// if a task has dependency and its parent task is inlined, then inline it (temporary fix)
	// (parent_task->inlined && !func_type->no_output_dep)
  if ( ((taskq->rtaskq_lp->scnt + taskq->rtaskq_lp->pcnt) > taskq->runtime_cutoff) &&
        (func_type->no_output_dep || (parent_task && (parent_task->sync_phase==0))) )
	{

		//if ((++rcutoff_count%10)==0) { UPC_POLL_NETWORK; }
		if ((++rcutoff_count%2)==0) { UPC_POLL_NETWORK; }
	
		DEBUG3( fprintf(dbg_fp, "[serialize] Inline start (taskID=%lu syncID=%lu)\n", 
						task_p->taskID, task_p->syncID); )
	
		taskq->task_stack_top++;
		//task_p->inlined = 1;
		if ( taskq->task_stack_top >= TASK_STACK_SZ) {
			printf("[taskq_put] OVERFLOW ERROR: MYTHREAD=%d task_stack_top=%d\n", 
							MYTHREAD, taskq->task_stack_top);
			upc_global_exit(1);
		}
		dst = (void *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top-1)*taskq->task_sz);
	
		memcpy(dst, task_p, sizeof(task_header_t));		// copy only the header (no input data copied)

#ifdef ENABLE_TIMER
		timer_stop(PUT_TIME);
#endif
	
		DEBUG3( fprintf(dbg_fp, "[serialize] taskID=%lu syncID=%lu task_stack_top=%d\n", 
						task_p->taskID, task_p->syncID, taskq->task_stack_top); )
		task_p->func(task_p->iptr, task_p->optr);		// has no wait_task and if it is my task
			
		DEBUG3( fprintf(dbg_fp, "[serialize] Inline done (taskID=%lu syncID=%lu)\n", 
						task_p->taskID, task_p->syncID); )

		taskq->task_stack_top--;
	
		ret = 0;
	}
	else {
		ret = taskq_insert(taskq, task_p, func_type);

#ifdef ENABLE_TIMER
		timer_stop(PUT_TIME);
#endif

	}
#else
	ret = taskq_insert(taskq, task_p, func_type);

#ifdef ENABLE_TIMER
		timer_stop(PUT_TIME);
#endif

#endif

	DEBUG3( fprintf(dbg_fp, "[taskq_put] done returns %d: taskID=%lu syncID=%lu\n", 
					ret, task_p->taskID, task_p->syncID); )

	return ret;
}

int taskq_insert(taskq_handle taskq, task_header_t *task_p, func_type_t *func_type)
{
	int canary2 = 67890;
	char *dst;
	int out_offs = 0;
	sync_node_t *prev_sync_node;
	rtaskq_t *rtaskq = taskq->rtaskq_lp;
	int prev_ptop = rtaskq->prev_ptop;
	int ptop = rtaskq->ptop;
	task_header_t *rtask_p;
	void *iptr = task_p->iptr;
	task_header_t *parent_task = NULL;

	if (taskq->task_stack_top > 0) {
		parent_task = (task_header_t *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top-1)*taskq->task_sz);
	}

	DEBUG3(	
		fprintf(dbg_fp, "[taskq_insert] taskID=%lu strt ", task_p->taskID);
		if (parent_task == NULL) 
			fprintf(dbg_fp, "(ptop=%d prev_ptop=%d)\n", ptop, prev_ptop); 
		else
			fprintf(dbg_fp, "(sync_phase=%d ptop=%d prev_ptop=%d)\n", parent_task->sync_phase, ptop, prev_ptop); 
	)

	// Optimize 1: do not copy data twice - one to task and the other to rtaskq
	// Optimize 2: if it is from taskq_put_continuation, do not copy
	if (func_type->iSize > 0) {		// if there is a data to copy
		dst = (char *)task_p + sizeof(task_header_t);
		memcpy(dst, iptr, func_type->iSize);
	}

	// if a root task (parent_task==NULL) or inserting a task before any taskq_fence
	if (parent_task==NULL || (parent_task && parent_task->sync_phase==0)) { 	
		rtaskq_push(taskq, task_p, taskq->task_sz); 	
	}
	else {		// multiple sync_phase has not been tested!
		if (prev_ptop == ptop) {	// no tasks has been put in the previous sync_phase
			rtaskq_push(taskq, task_p, taskq->task_sz); 	
		}
		else if (prev_ptop < ptop) {
#ifdef KDEBUG1
			// redirects the output of the depending tasks in the ready queue to the task
			// which is inserted into the wait queue.
			task_p = upc_map_put(taskq->wait_task_map, task_p->taskID, task_p);

/*
			printf("[taskq_insert] why am I here?\n");
			upc_global_exit(1);
*/

			while (prev_ptop < ptop) {
				rtask_p = (task_header_t *)(rtaskq->gbuf_lp + (prev_ptop * taskq->task_sz));
				// if the output of the previous tasks belongs to the input of the current task,
				// redirects the output pointer to the input address of the current task 
				if ( ((char *)rtask_p->optr >= (char *)iptr) && 
						 ((char *)rtask_p->optr < ((char *)iptr+func_type->iSize)) )
				{
					rtask_p->out_offs = ((char *)rtask_p->optr - (char *)iptr);
					// re-direct output to the output buffer in the wait queue
					rtask_p->optr = task_p + sizeof(task_header_t);
				}
				prev_ptop++;
			}

#else
			// update all the tasks pushed into the rtaskq during the previous phase
			while (prev_ptop < ptop) {
				rtask_p = (task_header_t *)(rtaskq->gbuf_lp + (prev_ptop * taskq->task_sz));
				// if the output of the previous tasks belongs to the input of the current task,
				// redirects the output pointer to the input address of the current task 
				if ( ((char *)rtask_p->optr >= (char *)iptr) && 
						 ((char *)rtask_p->optr < ((char *)iptr+func_type->iSize)) )
				{
					rtask_p->out_offs = ((char *)rtask_p->optr - (char *)iptr);
					DEBUG3( fprintf(dbg_fp,"[taskq_insert] taskID=%lu's output redirected (out_offs=%d)\n", 
									rtask_p->taskID, rtask_p->out_offs); )
				}
				prev_ptop++;
			}

			// note: re-assign the task pointer to the task address inserted in the wtaskq map 
			task_p = upc_map_put(taskq->wait_task_map, task_p->taskID, task_p);
#endif

			// update the address of the target wait task in the wtaskq for the previous sync phase
			DEBUG3(
				fprintf(dbg_fp, "[taskq_insert] parent_task: [child_syncID=%lu prev_child_syncID=%lu]\n", 
					parent_task->child_syncID, parent_task->prev_child_syncID); 
			)
			prev_sync_node = upc_map_get(taskq->sync_node_map, parent_task->prev_child_syncID);
			task_p->next = prev_sync_node->wait_task;		// set the next to the previous wait_task
			prev_sync_node->wait_task = task_p;					// then update the wait_task
		}
		else {
			printf("[taskq_insert] ERROR: prev_ptop(%d) > ptop(%d)\n", prev_ptop, ptop);
			upc_global_exit(1);
		}
	}

	if (parent_task) {
		parent_task->sync_tcnt++;	// increment the number tasks created in the current sync phase
	}
	
	DEBUG3(	
		fprintf(dbg_fp, "[taskq_insert] taskID=%lu syncID=%lu done\n", task_p->taskID, task_p->syncID);
	)

	return task_p->taskID;
}

void rtaskq_push(taskq_t *taskq, task_header_t *task, int task_sz)
{
	char *dst;
	rtaskq_t *rtaskq = taskq->rtaskq_lp;

	DEBUG3( 
		fprintf(dbg_fp, "[rtaskq_push] taskID=%lu task=%p task_sz=%d\n", task->taskID, task, task_sz); 
	)

	dst = (char *)rtaskq->gbuf_lp + (rtaskq->ptop * task_sz);

	memcpy(dst, task, task_sz);									// copy both task header and input

	rtaskq->ptop = (++rtaskq->ptop)%rtaskq->size;
	rtaskq->pcnt++;
	rtaskq->nPut++;

	if (rtaskq->pcnt + rtaskq->scnt >= READYQ_SZ) {
		printf("[rtaskq_push] ERROR: pid%d's rtaskq overflow! (pcnt=%d scnt=%d > %d)\n", MYTHREAD,
						rtaskq->pcnt, rtaskq->scnt, READYQ_SZ);
		upc_global_exit(1);
	}

	// check if the rtaskq has enough tasks to release
	if (rtaskq->pcnt > taskq->threshold) {
		taskq_release(taskq);
	}
}

/* msgbuf contains the result of the predecessor tasks from other UPC threads. 
	 Move these outputs to the paramenter list of the corresponding successor task */

int taskq_process_msgbuf(taskq_t *taskq) 
{
	int i;
	void *src, *dst;
	const char *header = "taskq_process_msgbuf";
	task_msgbuf_t *msgbuf_lp = taskq->msgbuf_lp;
	task_header_t *task_p, *wait_task;
	task_msg_header_t *msg_p;
	sync_node_t *sync_node;
	int msg_cnt = 0;
	
	DEBUG2( fprintf(dbg_fp, "[%s] strt: msgbuf_cnt=%d\n", header, msgbuf_cnt[MYTHREAD]); )

	upc_lock(msgbuf_lp->lock);	/* msgbuf access should always be locked */
	{
		msg_cnt = msgbuf_cnt[MYTHREAD]; 	// localizing shared var inside critical section should be ok.
		for (i=0; i < msg_cnt; i++) {
			msg_p = (task_msg_header_t *)((char *)msgbuf_lp->sbuf_lp + taskq->msg_sz*i);
			DEBUG3(fprintf(dbg_fp, "[%s] msg_p->syncID=%lu\n", header, msg_p->syncID); )
			sync_node = upc_map_get(taskq->sync_node_map, msg_p->syncID);
			if (sync_node != NULL) {		
				wait_task = sync_node->wait_task;
				if (msg_p->out_sz > 0) {
					src = (char *)msg_p + sizeof(task_msg_header_t);
					if (wait_task != NULL) {
						DEBUG3(fprintf(dbg_fp, "[%s] wait_task->taskID=%lu\n", header, wait_task->taskID); )
						dst = (char *)wait_task + sizeof(task_header_t) + msg_p->offset;
						DEBUG2( fprintf(dbg_fp, "[%s] i=%d src=%p, dst=%p\n", header, i, src, dst); )
						memcpy(dst, src, msg_p->out_sz);		
					}
					else {
						DEBUG2( fprintf(dbg_fp, "[%s] i=%d src=%p, optr=%p\n", header, i, src, msg_p->optr); )
						DEBUG2( fprintf(dbg_fp, "[%s] *src=%lu\n", header, *(uint64_t *)src); )
						memcpy(msg_p->optr, src, msg_p->out_sz);		
					}
				}

				DEBUG3(fprintf(dbg_fp, "[%s] sync_node->ref_count=%d\n", header, sync_node->ref_count); )
				if ( (--sync_node->ref_count) == 0 ) {
					while (wait_task != NULL) {
						rtaskq_push(taskq, wait_task, taskq->task_sz);
						// remove the wait_task
						upc_map_remove(taskq->wait_task_map, wait_task->taskID);	
						DEBUG2( fprintf(dbg_fp, "[%s] taskID=%lu task is now ready\n", header, wait_task->taskID); )
						wait_task = wait_task->next;
					}
					// remove the sync_node
					if (upc_map_remove(taskq->sync_node_map, msg_p->syncID) == NULL) {
						fprintf(dbg_fp, "[upc_map_remove] returns NULL for msg_p->syncID=%lu\n", msg_p->syncID);
						upc_global_exit(1);
					}
				}
			}
		}
		msgbuf_cnt[MYTHREAD] = 0;		// reset the msgbuf_cnt
	} 
	upc_unlock(msgbuf_lp->lock);

	return msg_cnt;
}

/* This function should not access parent_task because parent_task might have been expired */
int taskq_execute(taskq_t *taskq)
{
	void (*func0)();
	void (*func1)(void *);
	void (*func2)(void *, void *);
	task_header_t *task, *wait_task;
	void *idata_ptr, *odata_ptr;
	func_type_t *func_type;
	char *task_out_buf_p;
	int *my_msgbuf_cnt;
	char *dst;
	void *optr;
	sync_node_t *sync_node_p;
	
	DEBUG3( 
		fprintf(dbg_fp, "[taskq_execute] strt pcnt=%d stack_top=%d ------------------------------\n", 
		taskq->rtaskq_lp->pcnt, taskq->task_stack_top); 
	)

	// we assume that taskq_get will be called only from here
	// taskq_get returns task space from task_get_buf[taskq->task_stack_top];
	// task_stack_top should be updated after taskq_get function call
	if ( (task = taskq_get(taskq)) == NULL ) return _FAIL;

	my_msgbuf_cnt = (int *)&msgbuf_cnt[MYTHREAD];		// local access
	if (*my_msgbuf_cnt >= (int)(taskq->msgbuf_lp->size*0.5)) {				// optimize it comparison
		taskq_process_msgbuf(taskq);		// flush the message buffer
	}

	if ( (++taskq->syncID) < 0 ) {
		printf("[taskq_execute] tid%d: syncID overflow! taskID=%lu\n", MYTHREAD, taskq->syncID);
		upc_global_exit(1);
	}

	taskq->task_stack_top++;

	if ( taskq->task_stack_top >= TASK_STACK_SZ) {
		printf("OVERFLOW ERROR: MYTHREAD=%d task_stack_top=%d\n", MYTHREAD, taskq->task_stack_top);
		printf("Fix this overflow problem with dynamic re-sizing!\n");
		upc_global_exit(1);
	}

	// set prev_ptop to ptop to track which tasks will be pushed during the phase
	taskq->rtaskq_lp->prev_ptop = taskq->rtaskq_lp->ptop;

	DEBUG3 (
		fprintf(dbg_fp, "> taskID=%lu, syncID=%lu ", task->taskID, task->syncID);
		fprintf(dbg_fp, "stack_top=%d owner=%d task=%p optr=%p\n", taskq->task_stack_top, task->owner, task, task->optr);
	)

#ifdef DEBUG_STACK
	if (taskq->task_stack_top > taskq->task_stack_max) taskq->task_stack_max = taskq->task_stack_top;
	DEBUG3( fprintf(dbg_fp, " task_stack_max=%d\n", taskq->task_stack_max); )
#endif
	
	// -----------------------------------------------------------------------------
	// now, execute the task function
	// -----------------------------------------------------------------------------
	{
		func_type = &taskq->task_func_list[task->funcID];
		idata_ptr = (char *)task + sizeof(task_header_t);

		if (task->owner==MYTHREAD )
			optr = task->optr;
		else {
			// this is when the current thread executes a task that it stole.
			// the following output buffer pointer should be set after its index, task_stack_top, is updated
			// we can pre-allocate a heap memory for this because it is not that time critical case
			// optimize: we can remove this if stolen task's output can be copied back to
			// the owner thread by the notify_to_dep_tasks immediately such that there is no
			// stolen tasks executed before it calls notify_to_dep_tasks --> wait, there is a problem 
			// because user task function can call taskq_wait and taskq_wait can execute other stolen tasks
			optr = (char *) &taskq->task_out_buf[taskq->task_stack_top];	// optr is NULL if max_oSize is 0
		}

		if ((func_type->iSize > 0) && (func_type->oSize > 0)) {
			func2 = func_type->funcP;
			func2(idata_ptr, optr);
		}
		else if ((func_type->iSize > 0) && (func_type->oSize == 0)) {
			func1 = (void (*)(void *))func_type->funcP;
			func1(idata_ptr);
		}
		else if ((func_type->iSize == 0) && (func_type->oSize > 0)) {
			func1 = (void (*)(void *))func_type->funcP;
			func1(optr);
		}
		else {
			func0 = func_type->funcP;
			func0();
		}
	}

	// - sync_phase is incremented by taskq_fence and taskq_wait reset it to zero
	// - (syncID and sync_node) pair is added by taskq_fence and taskq_wait
	// - no sync_node for continuation task 
	// - no sync_node for a task if its parent task has neither taskq_fence nor taskq_wait

	DEBUG3 (
		fprintf(dbg_fp, "< taskID=%lu, syncID=%lu, ", task->taskID, task->syncID);
		fprintf(dbg_fp, "stack_top=%d, owner=%d, task=%p\n", taskq->task_stack_top, task->owner, task);
	)

	// do nothing if the current task has a continuation task, which will do the work
	// or there is no task that is dependent on the current task 

	// --------------------------------------------------------------------
	// Notify output to dependent tasks if the current task does not have 
	// a continuation task as a child AND has an output dependency
	// --------------------------------------------------------------------
	if (!task->hasCont && !func_type->no_output_dep) {	
	
		if (task->owner == MYTHREAD) {

			sync_node_p = upc_map_get(taskq->sync_node_map, task->syncID);

			if (sync_node_p != NULL) {
				wait_task = sync_node_p->wait_task;
	
/*
#ifdef KDEBUG1
#else
*/
				if (task->out_size > 0) {
					if (wait_task != NULL) {
						dst = (char *)wait_task + sizeof(task_header_t) + task->out_offs;
						memcpy(dst, optr, task->out_size);		
						DEBUG3 ( fprintf(dbg_fp, "[taskq_execute] dst=%p, *(int *)dst=%d\n", dst, *(int *)dst); )
					}
					else {
						// do nothing - the output has been written directly to task->optr
					}
				}
/*
#endif
*/
	
				if ( (--sync_node_p->ref_count) == 0 ) {
					while ( wait_task != NULL ) {
						rtaskq_push(taskq, wait_task, taskq->task_sz);
						// remove the wait_task
						upc_map_remove(taskq->wait_task_map, wait_task->taskID);	
						wait_task = wait_task->next;
					}
					// remove the sync_node
					//upc_map_remove(taskq->sync_node_map, task->syncID);	
					if (upc_map_remove(taskq->sync_node_map, task->syncID) == NULL) {
						fprintf(dbg_fp, "[upc_map_remove] returns NULL for msg_p->syncID=%lu\n", task->syncID);
						upc_global_exit(1);
					}
				}
			}
		}
		else {
			int tid = task->owner;
			shared void *remote_msg_p;
			shared task_msgbuf_t *remote_msgbuf_sp = &taskq->msgbuf_sp[tid];
			task_msg_header_t *src_msg_p = taskq->task_msg_buf;
			src_msg_p->syncID = task->syncID;
			src_msg_p->out_sz = task->out_size;
			src_msg_p->offset = task->out_offs;
			src_msg_p->optr = task->optr;
	
			// copy output data into the task msg; (char *) is critical
			if (task->out_size > 0) {
				void *src = (void *)(optr);
				void *dst = (void *)(((char *)(src_msg_p))+sizeof(task_msg_header_t));
				memcpy(dst, src, task->out_size);	
				DEBUG3( fprintf(dbg_fp, "[execute] send *src=%d\n", *(int *)src); )
			}
	
			DEBUG3( fprintf(dbg_fp, "[execute] send task(taskID=%lu)'s msg(syncID=%lu) to tid%d\n", 
							task->taskID, src_msg_p->syncID, tid); )
	
			upc_lock(remote_msgbuf_sp->lock);
			{
				int msg_cnt = msgbuf_cnt[tid];		// make a copy of the shared variable
				if (msg_cnt >= taskq->msgbuf_lp->size) {	/* all threads have the same msgbuf size */
					printf("[notify_to_dep_tasks] tid%d: tid%d's msgbuf overflow! (msgbuf_cnt=%d)\n", 
									MYTHREAD, tid, msg_cnt);
					upc_global_exit(1);
				}
				remote_msg_p = (shared void *)(remote_msgbuf_sp->sbuf + THREADS*taskq->msg_sz*msg_cnt);
				upc_memput(remote_msg_p, src_msg_p, sizeof(task_msg_header_t) + task->out_size);
				//remote_msgbuf_sp->msg_cnt++;
				msgbuf_cnt[tid]++;
			}
			upc_unlock(remote_msgbuf_sp->lock);
		}
	}

	// finalize routine for taskq_execute
	taskq->rtaskq_lp->prev_ptop = taskq->rtaskq_lp->ptop;		// update prev_ptop

	DEBUG3( fprintf(dbg_fp, "[taskq_execute] taskID=%lu stack=%d done --------------------\n", 
					task->taskID, taskq->task_stack_top); )

	taskq->task_stack_top--;

	DEBUG3( fprintf(dbg_fp, "[taskq_execute] taskID=%lu done  ----------------\n",task->taskID);  )
	return _SUCCESS;
}

task_header_t *taskq_get(taskq_handle taskq)
{
	char *src;
	task_header_t *task = (task_header_t *)((char *)(&taskq->task_get_buf[0]) + (taskq->task_stack_top)*taskq->task_sz);
	rtaskq_t *rtaskq = taskq->rtaskq_lp;
	int size = rtaskq->size;

/*
	fprintf(dbg_fp, "[taskq_get] task=%p, &taskq->task_get_buf[0]=%p, taskq->task_stack_top=%d taskq->task_sz=%d\n",  task, &taskq->task_get_buf[0], taskq->task_stack_top, taskq->task_sz);
*/

	DEBUG3 ( 
		fprintf(dbg_fp, "[taskq_get] strt: pcnt=%d ------------------------------------\n", rtaskq->pcnt);
	)

	/* when the ready queue is empty */
	if (rtaskq->pcnt == 0) {
		if ( !taskq_acquire(taskq) ) { 
			taskq_local_reset(taskq);
			return NULL; 		// if failed to re-acquire, then return NULL
		}
	}

	rtaskq->ptop = (rtaskq->ptop-1+size)%size;		/* decrement ptop by one */
	src = (void *)(rtaskq->gbuf_lp + (rtaskq->ptop * taskq->task_sz));

	DEBUG3 (	// memcpy address check
		void *lb = (void *)(rtaskq->gbuf_lp);
		void *ub = (void *)(rtaskq->gbuf_lp + rtaskq->size*taskq->task_sz);
		if (src < (char *)lb || src >= (char *)ub) {
			fprintf(dbg_fp, "  rtaskq->gbuf=[%p:%p] ", lb, ub);
			fprintf(dbg_fp, "src=%p, taskq->task_sz=%d\n", src, taskq->task_sz);
		}
	)

	// copy a task + data   (note: to-do -> separate task and data - copy the actual size )
	memcpy(task, src, taskq->task_sz);

	rtaskq->pcnt--;
	rtaskq->nGet++;

	DEBUG3 (
		fprintf(dbg_fp, "[taskq_get] done: task=%p (taskID=%lu, task_stack_top=%d) -------------\n\n", 
						task, task->taskID, taskq->task_stack_top);
	)

	return task;
}

void taskq_all_free(taskq_t *taskq)
{
	DEBUG1( fprintf(dbg_fp, "[taskq_free] freeing task queue memory\n"); )

	// free local memories
	upc_map_free(taskq->wait_task_map);

	// free UPC shared memories
	if (MYTHREAD==0) {		/* root thread has the saddr of the shared memory */
		upc_free(taskq->rtaskq_sp->gbuf);
		upc_free(taskq->msgbuf_sp->sbuf);
		upc_free(taskq->rtaskq_sp);
		upc_free(taskq->msgbuf_sp);
	}
}

int taskq_release(taskq_t *taskq)
{
	int i;
	int taskReleased = 0;
	char header[24] = "taskq_release";
	rtaskq_t *rtaskq = taskq->rtaskq_lp;
	int k = taskq->chunk_size;
	func_type_t *func_type;

	DEBUG3 ( 
		fprintf(dbg_fp, "[%s] k=%d scnt=%d pcnt=%d\n", header, k, rtaskq->scnt, rtaskq->pcnt); 
	)

	if (rtaskq->pcnt < k) {
    fprintf(dbg_fp, "[taskq_release] returns 0 (pcnt=%d < k=%d)\n", rtaskq->pcnt, k);
    return 0;
	}

	upc_lock(rtaskq->lock);
	{
		rtaskq->pbot = (rtaskq->pbot + k) % rtaskq->size;
		rtaskq->scnt += k;
		rtaskq->pcnt -= k;
		taskReleased = k;
		rtaskq->nRelease += k;
	}
	upc_unlock(rtaskq->lock);

	DEBUG3 (
		if (taskReleased==0) {
			fprintf(dbg_fp, "[taskq_release] failed to release tasks\n");
		}
		fprintf(dbg_fp, "[taskq_release] taskReleased=%d\n", taskReleased); 
	)
	return taskReleased;
}

int taskq_acquire(taskq_t *taskq)
{
	int k;
	int max_acquire = taskq->chunk_size;
	int taskAcquired = 0;
	rtaskq_t *rtaskq = taskq->rtaskq_lp; 
	
	DEBUG3( fprintf(dbg_fp, "[taskq_acquire] max_acquire=%d\n", max_acquire); )

	upc_lock(rtaskq->lock);
	{
		k = rtaskq->scnt;
		if ( k > 0 ) {
			if (k > max_acquire) k = max_acquire;
			rtaskq->pbot = (rtaskq->pbot - k + rtaskq->size) % rtaskq->size;
			rtaskq->pcnt += k;
			rtaskq->scnt -= k;
			rtaskq->nAcquire += k;	/* k should be a multiple of chunk */
			taskAcquired = k;
		}
		DEBUG3( fprintf(dbg_fp, "[taskq_acquire] rtaskq->lock=%lu\n", rtaskq->lock); )
	}
	upc_unlock(rtaskq->lock);

	DEBUG3( fprintf(dbg_fp, "[taskq_acquire] taskAcquired=%d\n", taskAcquired); )

	return taskAcquired;
}

int is_local_threads(int tid) {

	if (num_local_threads==0) {
		taskq_error_exit("[is_local_threads] needs taskq_set_core_per_node call for non-BUPC\n");
	}

	for (int i=0; i<num_local_threads; i++) {
		if (tid == local_thread_list[i]) return 1;
	}
	return 0;
}

int taskq_find_victim(taskq_t *taskq, int k, int first_trial)
{
	int trial_cnt;
	int i, candidate, offset, bottom, victim_work;
	int steal_mode = taskq->steal_mode;
	int chunk_mode = taskq->chunk_mode;
	int steal_from_the_same_node; 
	float scale;
	char header[24] = "taskq_find_victim";
	shared rtaskq_t *vt_taskq;
	shared rtaskq_t *rtaskq_sp = taskq->rtaskq_sp;
	//static int toggle=0;

	DEBUG_STEAL( fprintf(dbg_fp, "[%s] k=%d strt\n", header, k); )

	//toggle = (toggle+1)%2;
	// conditions for random victim selection: 
	//   case 1: every steal trial in RANDOM
	//   case 2: until the first successful steal in INTRA_NODE_ONLY
	//   case 3: after the preset trials in LAWS_RAND or LAWS_NEXT
	for (trial_cnt=0; trial_cnt < taskq->total_steal_trial; trial_cnt++) {

		candidate = -1;
		if (steal_mode==HIER_WORK_STEAL || (steal_mode==SAME_NODE_STEAL && taskq->rtaskq_lp->nSteal>0)) {
			candidate = taskq->victim_list[trial_cnt];
		}

		if (taskq->first_steal_from_root && (MYTHREAD!=0)) {
      candidate = -1;
    }

		if (candidate < 0) {
			do {
				scale = rand()/(float)RAND_MAX;   /* scale = [0:1] */
				candidate = (int)(scale*THREADS);
			} while (candidate == MYTHREAD);
		}

		vt_taskq = (shared rtaskq_t *) &rtaskq_sp[candidate];
		victim_work = vt_taskq->scnt;

		DEBUG_STEAL(
			fprintf(dbg_fp, "  trial_cnt=%d, candidate=%d, work=%d\n", trial_cnt, candidate, victim_work);
		)

		steal_from_the_same_node = is_local_threads(candidate);
		if ( chunk_mode==HALF_CHUNK || ( chunk_mode==ADAPTIVE_CHUNK && steal_from_the_same_node==0 ) ) {
			k = (int)(victim_work/2);
			if (k==0) k = 1;
		}

		if (victim_work >= k) {
			DEBUG_STEAL(
				fprintf(dbg_fp, "[%s] found %d tasks from victim=%d\n", header, victim_work, candidate); 
			)
			return candidate;
		}
		else {
			taskq->rtaskq_lp->failed_victim_probe++;	
		}
	}
	DEBUG_STEAL( fprintf(dbg_fp, "[%s] finds no work (k=%d)\n", header, k); )
	return NOT_FOUND;
}

int taskq_steal(taskq_t *taskq)
{
	int taskStealed = 0;
	int victimID;
	int steal_size = taskq->steal_size;
	int first_trial = 1;

	DEBUG2( fprintf(dbg_fp, "[taskq_steal] strt (steal_size=%d)\n", steal_size); )

	if (THREADS < 2 || steal_size==0) return 0;

	while ( (victimID = taskq_find_victim(taskq, taskq->steal_size, first_trial)) != NOT_FOUND ) {
		//fprintf(dbg_fp, "[taskq_steal] victimID=%d\n", victimID);
		if ( (taskStealed = taskq_steal_engine(taskq, victimID, steal_size)) > 0 ) {
			taskq->first_steal_from_root = 0;
			break;
		}
		first_trial = 0;
	}

	DEBUG3(
		if (taskStealed>0)
			fprintf(dbg_fp, "[taskq_steal] stealed %d tasks from victim=%d\n", taskStealed, victimID);
		else
			fprintf(dbg_fp, "[taskq_steal] failed\n");
	)

	return taskStealed;
}

int taskq_steal_engine(taskq_t *taskq, int vt_tid, int numSteal)
{
	shared const void *src;
	const void *src_lp;
	void *dst;
	int taskStealed = 0;
	int steal_from_the_same_node; 
	int chunk_mode = taskq->chunk_mode;
	char header[24] = "[taskq_steal_engine]";
	rtaskq_t *rtaskq = taskq->rtaskq_lp;
	shared rtaskq_t *vt_taskq;
	rtaskq_t *vt_taskq_lp;
		
	DEBUG_STEAL( 
		fprintf(dbg_fp, "%s start: thief=%d victim=%d numSteal=%d\n", header, MYTHREAD, vt_tid, numSteal);
	)
	if ((rtaskq->pcnt + rtaskq->scnt + numSteal) > rtaskq->size) {
		fprintf(dbg_fp, "%s ERROR: pcnt=%d, scnt=%d\n", header, rtaskq->pcnt, rtaskq->scnt);
		taskq_error_exit("[taskq_steal] No need to steal");
	}

	if (rtaskq->ptop + numSteal > rtaskq->size) {
		if (rtaskq->pcnt == 0 && rtaskq->scnt == 0) {
			taskq_local_reset(taskq);
		}
		else {
			fprintf(dbg_fp, "%s ERROR: scnt=%d, pcnt=%d ", header, rtaskq->scnt, rtaskq->pcnt); 
			fprintf(dbg_fp, "ptop=%d pbot=%d tail=%d\n", rtaskq->ptop, rtaskq->pbot, rtaskq->tail);
			taskq_error_exit("[taskq_steal_engine] is not empty!");
		}
	}

	vt_taskq = (shared rtaskq_t *) &(taskq->rtaskq_sp[vt_tid]);
	steal_from_the_same_node = is_local_threads(vt_tid);
	upc_lock(vt_taskq->lock);
	{
		DEBUG3( fprintf(dbg_fp, "%s vt=%lu's {pbot=%d, tail=%d, scnt=%d}\n", header, 
							upc_threadof(vt_taskq), vt_taskq->pbot, vt_taskq->tail, vt_taskq->scnt); ) 

		if ( chunk_mode==HALF_CHUNK || (chunk_mode==ADAPTIVE_CHUNK && steal_from_the_same_node==0) ) {
			numSteal = (int)(vt_taskq->scnt/2);
			if (numSteal==0) numSteal = 1;
		}
			
		if (vt_taskq->scnt >= numSteal) {

			if (vt_taskq->tail + numSteal > rtaskq->size) {
				numSteal = rtaskq->size - vt_taskq->tail;		// adjust the number of task to steal
			}

			src = (shared const void *)(vt_taskq->gbuf + (THREADS*vt_taskq->tail*taskq->task_sz));
			dst = (void *) ((char *)rtaskq->gbuf_lp + (rtaskq->ptop * taskq->task_sz));
			//DEBUG3( fprintf(dbg_fp, "%s dst=%p, src=%lu, numSteal*taskq->task_sz=%d\n", header, dst, 
			//				src, numSteal*taskq->task_sz); )
			upc_memget(dst, src, numSteal*taskq->task_sz);

			DEBUG1( fprintf(dbg_fp, "SUCCESS: stealed %d tasks from victim=%d(scnt=%d)\n", 
							numSteal, vt_tid, vt_taskq->scnt); )

			vt_taskq->tail = (vt_taskq->tail + numSteal) % rtaskq->size;
			vt_taskq->scnt = (vt_taskq->scnt - numSteal + rtaskq->size) % rtaskq->size;
			taskStealed = numSteal;
		}
	}
	upc_unlock(vt_taskq->lock);	

	if (taskStealed > 0) {
		rtaskq->ptop = (rtaskq->ptop + taskStealed) % rtaskq->size;
		rtaskq->pcnt = (rtaskq->pcnt + taskStealed) % rtaskq->size ;
		rtaskq->nTaskSteal += taskStealed;
		rtaskq->nSteal++;
		victim_count[vt_tid]++;
		DEBUG_STEAL( 
			fprintf(dbg_fp, "%s SUCCESS: stealed %d tasks from victim=%d\n", header, taskStealed, vt_tid);
		)
	}
	else {
		taskq->rtaskq_lp->failed_steal_attempt++;	
		DEBUG_STEAL( 
			fprintf(dbg_fp, "%s FAIL: no %d tasks from victim=%d\n", header, taskStealed, vt_tid);
		)
	}

	return taskStealed;
}

void taskq_check(taskq_t *taskq)
{
	rtaskq_t *rtaskq_lp = taskq->rtaskq_lp;

/*
	printf("[taskq_check] nPut[%d]=%d nGet[%d]=%d\n", MYTHREAD, nPut, nGet);
	if (MYTHREAD==0) { total_nPut = 0; total_nGet=0; }
*/
	rtaskq_lp->nProduced[MYTHREAD] = rtaskq_lp->nPut;
	rtaskq_lp->nConsumed[MYTHREAD] = rtaskq_lp->nGet;

	upc_barrier;

/*
	upc_all_reduceI(&total_nPut, &rtaskq_lp->nProduced[MYTHREAD], UPC_ADD, THREADS, 1, NULL, 0);
	upc_all_reduceI(&total_nGet, &rtaskq_lp->nConsumed[MYTHREAD], UPC_ADD, THREADS, 1, NULL, 0);
*/
	fprintf(dbg_fp, "\n------------------- taskq_check --------------------\n");

	if (MYTHREAD==0) {
		int total_nPut=0, total_nGet=0, total_nSteal=0, total_nTaskSteal=0;
		printf("------------------- taskq_check --------------------\n");
		for (int i=0; i<THREADS; i++) {
			shared rtaskq_t *rtaskq_sp = &taskq->rtaskq_sp[i];
			printf("tid=%d, nPut=%d, nGet=%d, nSteal=%d, nTaskSteal=%d\n", i, rtaskq_sp->nPut, 
							rtaskq_sp->nGet, rtaskq_sp->nSteal, rtaskq_sp->nTaskSteal);
			fprintf(dbg_fp, "tid=%d, nPut=%d, nGet=%d, nSteal=%d, nTaskSteal=%d\n", i, rtaskq_sp->nPut, 
							rtaskq_sp->nGet, rtaskq_sp->nSteal, rtaskq_sp->nTaskSteal);
										
			total_nPut += rtaskq_sp->nPut; 
			total_nGet += rtaskq_sp->nGet; 
			total_nSteal += rtaskq_sp->nSteal; 
			total_nTaskSteal += rtaskq_sp->nTaskSteal; 
		}
		printf("\n-- total_nPut = %d, total_nGet = %d\n", total_nPut, total_nGet);
		printf("-- total_nSteal = %d totalTaskSteal = %d\n", total_nSteal, total_nTaskSteal);
		printf("-- wait_task_map has %d entries left\n", taskq->wait_task_map->entry_cnt);
		printf("-- sync_node_map has %d entries left\n", taskq->sync_node_map->entry_cnt);
		fprintf(dbg_fp, "-- total_nPut = %d, total_nGet = %d\n", total_nPut, total_nGet);
		fprintf(dbg_fp, "-- total_nSteal = %d total_nTaskSteal=%d\n", total_nSteal, total_nTaskSteal);

		if (total_nPut == total_nGet) {
			printf("-- check: equal total_nPut(%d) == total_nGet(%d)\n", total_nPut, total_nGet);
			fprintf(dbg_fp, "check: equal total_nPut(%d) == total_nGet(%d)\n", total_nPut, total_nGet);
		}
		else {
			printf("ERROR: total: total_nPut(%d) != total_nGet(%d)\n", total_nPut, total_nGet);
			fprintf(dbg_fp, "ERROR: total: total_nPut(%d) != total_nGet(%d)\n", 
							total_nPut, total_nGet);
			upc_global_exit(1);
		}
		printf("\n");
	}

	fprintf(dbg_fp, "-- wait_task_map has %d entries left\n", taskq->wait_task_map->entry_cnt);
	fprintf(dbg_fp, "-- sync_node_map has %d entries left\n", taskq->sync_node_map->entry_cnt);
	fprintf(dbg_fp, "----------------------------------------------------\n");

}

int taskq_local_depth(taskq_t *taskq)
{
	return taskq->rtaskq_lp->pcnt;
}

void taskq_snapshot(taskq_t *taskq)
{
	int i, j;
	char *task;
	rtaskq_t *rtaskq = taskq->rtaskq_lp;
	
	i = rtaskq->ptop;
	j = 0;
	fprintf(dbg_fp, "\n-----------------------------------------------------------------\n");
	fprintf(dbg_fp, "- size=%d pcnt=%d scnt=%d ", rtaskq->size, rtaskq->pcnt, rtaskq->scnt);
	fprintf(dbg_fp, "task_sz=%d max_iSize=%d mst_sz=%d\n", taskq->task_sz, taskq->max_iSize, taskq->msg_sz);
	fprintf(dbg_fp, "- ptop=%d pbot=%d tail=%d ", rtaskq->ptop, rtaskq->pbot, rtaskq->tail);
	fprintf(dbg_fp, "prev_ptop=%d\n", rtaskq->prev_ptop);
	fprintf(dbg_fp, "-----------------------------------------------------------------\n");
	fprintf(dbg_fp, "1. Printing private task queue\n");
	while (j++ < rtaskq->pcnt) {
		i = (--i + rtaskq->size) % rtaskq->size;
		task =  rtaskq->gbuf_lp + (i * taskq->task_sz);
		fprintf(dbg_fp, "  pq[%d] = [task=%d]\n", i, *(int *)task);
	}
	fprintf(dbg_fp, "\n");
	fflush(dbg_fp);

	i = rtaskq->pbot;
	j = 0;
	fprintf(dbg_fp, "2. Printing shared task queue\n");
	while (j++ < rtaskq->scnt) {
		i = (--i + rtaskq->size) % rtaskq->size;
		task = rtaskq->gbuf_lp + (i * taskq->task_sz);
		fprintf(dbg_fp, "  sq[%d] = [task=%d]\n", i, *(int *)task);
	}
	fflush(dbg_fp);

	taskq_stats(taskq);
	fprintf(dbg_fp, "\n");
}

void taskq_remote_snapshot(taskq_t *taskq, int remote_tid)
{
	int i, j;
	shared task_header_t *task;
	shared rtaskq_t *rtaskq_sp = (shared rtaskq_t *) &(taskq->rtaskq_sp[remote_tid]);
	
	upc_lock(rtaskq_sp->lock);

	i = rtaskq_sp->ptop;
	j = 0;
	fprintf(dbg_fp, "\n----------------------  REMOTE QUEUE  ---------------------------\n");
	fprintf(dbg_fp, "- size=%d pcnt=%d scnt=%d\n", rtaskq_sp->size, rtaskq_sp->pcnt, rtaskq_sp->scnt);
	fprintf(dbg_fp, "- ptop=%d pbot=%d tail=%d\n", rtaskq_sp->ptop, rtaskq_sp->pbot, rtaskq_sp->tail);
	fprintf(dbg_fp, "-----------------------------------------------------------------\n");
	fprintf(dbg_fp, "1. Printing remote task queue (can be incorret)\n");
	while (j++ < rtaskq_sp->pcnt) {
		i = (--i + rtaskq_sp->size) % rtaskq_sp->size;
		task =  (shared task_header_t *)( rtaskq_sp->gbuf + THREADS*(i * taskq->task_sz) );
		fprintf(dbg_fp, "  pq[%d] = [task->taskID=%ld]\n", i, task->taskID);
	}
	fprintf(dbg_fp, "\n");
	fflush(dbg_fp);

	i = rtaskq_sp->pbot;
	j = 0;
	fprintf(dbg_fp, "2. Printing remote shared task queue\n");
	while (j++ < rtaskq_sp->scnt) {
		i = (--i + rtaskq_sp->size) % rtaskq_sp->size;
		task =  (shared task_header_t *)( rtaskq_sp->gbuf + THREADS*(i * taskq->task_sz) );
		fprintf(dbg_fp, "  sq[%d] = [task->taskID=%ld]\n", i, task->taskID);
	}
	fflush(dbg_fp);

	taskq_remote_stats(taskq, remote_tid);

	upc_unlock(rtaskq_sp->lock);
}

// a collective function
void taskq_stats(taskq_t *taskq)
{
	int i, total_rcutoff_count=0;
	int max_victim = 0;
	float victim_percent, max_victim_percent = 0.0;
	rtaskq_t *rtq = taskq->rtaskq_lp;
	rcutoff_counts[MYTHREAD] = rcutoff_count;

	upc_barrier;

	if (MYTHREAD==0) {
		printf("----------------------- STAT -----------------------\n");
		printf(" READYQ_SZ=%d, TASK_STACK_SZ=%d\n", READYQ_SZ, TASK_STACK_SZ);
		printf(" WAITQ_BUCKETS=%d, WAITQ_ENTRIES=%d, MSGBUF_SZ=%d\n", WAITQ_BUCKETS, WAITQ_ENTRIES, MSGBUF_SZ);
		printf(" steal_size=%d chunk_size=%d ", taskq->steal_size, taskq->chunk_size);
		printf(" threshold=%d\n total_steal_trial=%d ", taskq->threshold, taskq->total_steal_trial);
		printf(" preset_steal_trial=%d\n", taskq->preset_steal_trial);
		printf(" steal_mode=%d, chunk_mode=%d\n", taskq->steal_mode, taskq->chunk_mode);
		printf(" failed_victim_probe=%d", taskq->rtaskq_lp->failed_victim_probe);
		printf(" failed_steal_attempt=%d\n", taskq->rtaskq_lp->failed_steal_attempt);
		printf(" task_sz=%d, max_iSize=%d", taskq->task_sz, taskq->max_iSize);
		printf(" msg_sz=%d\n", taskq->msg_sz);
#ifdef RUNTIME_CUTOFF
		printf(" runtime_cutoff=%d\n", taskq->runtime_cutoff);
		for (i=0; i<THREADS; i++) { total_rcutoff_count += rcutoff_counts[i]; }
		printf(" rcutoff_count= %d, total_rcutoff_count=%d\n", rcutoff_count, total_rcutoff_count);
#else
		printf(" RUNTIME CUTOFF is NOT applied: rcutoff_count= %d\n", rcutoff_count);
#endif

#ifdef DEBUG_STACK
		printf(" task_stack_top=%d, ", taskq->task_stack_top);
		printf(" task_stack_max=%d\n", taskq->task_stack_max);
#else
#endif
		printf("----------------------------------------------------\n");
	}

#ifdef ANALYSIS
		fprintf(dbg_fp, "----------------- RTASKQ SNAPSHOT ------------------\n");
		fprintf(dbg_fp, " sample_count1 = %d sample_count2 = %d\n", sample_count1, sample_count2);
		fprintf(dbg_fp, "----------------------------------------------------\n");
		fprintf(dbg_fp, "  scnt     pcnt\n");
		int scnt_min = 1073741824;		// 2^30
		int scnt_max = 0;
		int pcnt_min = 1073741824;		// 2^30
		int pcnt_max = 0;
		float pcnt_avg=0.0, scnt_avg=0.0;
		float pcnt_var=0.0, scnt_var=0.0;
		for (int i=0; i<sample_count2; i++) {
			if (pcnt_min > pcnt_top[i]) pcnt_min = pcnt_top[i];
			if (pcnt_max < pcnt_top[i]) pcnt_max = pcnt_top[i];
			if (scnt_min > scnt_top[i]) scnt_min = scnt_top[i];
			if (scnt_max < scnt_top[i]) scnt_max = scnt_top[i];
			pcnt_avg += (float)pcnt_top[i];
			scnt_avg += (float)scnt_top[i];
			fprintf(dbg_fp, "%6d  %6d\n", scnt_top[i], pcnt_top[i]); 
		}
		pcnt_avg /= (float)sample_count2;
		scnt_avg /= (float)sample_count2;
		for (int i=0; i<sample_count2; i++) {
			pcnt_var += (pcnt_top[i] - pcnt_avg) * (pcnt_top[i] - pcnt_avg) / (float)sample_count2;
			scnt_var += (scnt_top[i] - scnt_avg) * (scnt_top[i] - scnt_avg) / (float)sample_count2;
		}
		fprintf(dbg_fp, "scnt: min=%d max=%5d avg=%10.3f var=%10.3f\n", scnt_min, scnt_max, 
						scnt_avg, scnt_var);
		fprintf(dbg_fp, "pcnt: min=%d max=%5d avg=%5.3f var=%10.3f\n", pcnt_min, pcnt_max, 
						pcnt_avg, pcnt_var);
#endif

	fprintf(dbg_fp, "----------------------- STAT -----------------------\n");
	fprintf(dbg_fp, " size=%d, scnt=%d, pcnt=%d", rtq->size, rtq->scnt, rtq->pcnt);
	fprintf(dbg_fp, " ptop=%d, pbot=%d, tail=%d\n", rtq->ptop, rtq->pbot, rtq->tail);
	fprintf(dbg_fp, " steal_size=%d chunk_size=%d", taskq->steal_size, taskq->chunk_size);
	fprintf(dbg_fp, " threshold=%d\n total_steal_trial=%d ", taskq->threshold, taskq->total_steal_trial);
	fprintf(dbg_fp, " preset_steal_trial=%d\n", taskq->preset_steal_trial);
	fprintf(dbg_fp, " task_sz=%d, max_iSize=%d", taskq->task_sz, taskq->max_iSize);
	fprintf(dbg_fp, " msg_sz=%d\n", taskq->msg_sz);
	fprintf(dbg_fp, " nAcquire=%d nRelease=%d\n", rtq->nAcquire, rtq->nRelease);
	fprintf(dbg_fp, " nSteal=%d, nTaskSteal=%d, nFail=%d\n", rtq->nSteal, rtq->nTaskSteal, rtq->nFail);
	fprintf(dbg_fp, " nPut=%d, nGet=%d, nWakeUp=%d\n", rtq->nPut, rtq->nGet, rtq->nWakeUp);
	fprintf(dbg_fp, " msgbuf's size=%d, ", taskq->msgbuf_lp->size);
	fprintf(dbg_fp, " failed_victim_probe=%d", rtq->failed_victim_probe);
	fprintf(dbg_fp, " failed_steal_attempt=%d\n", rtq->failed_steal_attempt);
#ifdef ENABLE_TIMER
	fprintf(dbg_fp, " taskq_put: total_time=%lf msec, count=%d, avg time=%lf usec\n", 
					1000*timer_read(PUT_TIME), rtq->nPut, (1000000*timer_read(PUT_TIME))/(double)rtq->nPut);
#endif

#ifdef RUNTIME_CUTOFF
	fprintf(dbg_fp, " rcutoff_count=%d runtime_cutoff=%d\n", rcutoff_count, taskq->runtime_cutoff);
#else
	fprintf(dbg_fp, " RUNTIME_CUTOFF is NOT applied: rcutoff_count=%d\n", rcutoff_count);
#endif
	if (rtq->nSteal>0) {
		fprintf(dbg_fp, "victim_count : # of successful steals from victims\n"); 
		for (i=0; i<THREADS; i++) {
			victim_percent = 100.0*victim_count[i]/(float)rtq->nSteal;
			fprintf(dbg_fp, "p%d: %d (%6.3f %%) ", i, victim_count[i], victim_percent);
			if (i%4 == 0) fprintf(dbg_fp, "\n");
			if (max_victim_percent < victim_percent) {
				max_victim = i;
				max_victim_percent = victim_percent;
			}
		}
	}
	fprintf(dbg_fp, "\n");
	fprintf(dbg_fp, "tid %d's the most victim: %d (%f %%)\n", MYTHREAD, max_victim, max_victim_percent);
#ifdef DEBUG_STACK
	fprintf(dbg_fp, " task_stack_top=%d, ", taskq->task_stack_top);
	fprintf(dbg_fp, " task_stack_max=%d\n", taskq->task_stack_max);
#endif
	fprintf(dbg_fp, "----------------------------------------------------\n");
	fflush(dbg_fp);
}

void taskq_remote_stats(taskq_t *taskq, int tid)
{
	shared rtaskq_t *rtaskq_sp = &taskq->rtaskq_sp[tid];
	fprintf(dbg_fp, "-------------------------------------------------\n");
	fprintf(dbg_fp, "-                  STAT                          \n");
	fprintf(dbg_fp, "-------------------------------------------------\n");
	fprintf(dbg_fp, " size=%d, scnt=%d, pcnt=%d\n", rtaskq_sp->size, rtaskq_sp->scnt, rtaskq_sp->pcnt);
	fprintf(dbg_fp, " ptop=%d, pbot=%d, tail=%d\n", rtaskq_sp->ptop, rtaskq_sp->pbot, rtaskq_sp->tail);
	fprintf(dbg_fp, " nAcquire=%d, nRelease=%d\n", rtaskq_sp->nAcquire, rtaskq_sp->nRelease);
	fprintf(dbg_fp, " nSteal=%d nTaskSteal=%d\n", rtaskq_sp->nSteal, rtaskq_sp->nTaskSteal);
	fprintf(dbg_fp, " nFail=%d\n", rtaskq_sp->nFail);
	fprintf(dbg_fp, "-------------------------------------------------\n");
	fflush(dbg_fp);
}


/***********************************************************
 *  Utility Functions                                      *
 ***********************************************************/

int getLog2Val(int n)
{
	int depth = 0;
	while (n>1) {
		n = n >> 1;
		depth++;
	}
	return depth;
}

