/***********************************************************
 *                                                         *
 *                  UPC TASK QUEUE Library                 *
 *                                                         *
 ***********************************************************/

#ifndef _TASKQ_H
#define	_TASKQ_H

#include <stdio.h>
#include <stdlib.h>

/***********************************************************
 *                  #define VARIABLES                      *
 ***********************************************************/

#define PRINT_STACK_DEPTH
#ifdef PRINT_STACK_DEPTH
#define STACK_INDENT print_stack_indent()
#else
#define STACK_INDENT 
#endif

#define DEBUG_LEVEL0
/*
#define DEBUG_LEVEL1
#define DEBUG_LEVEL2
#define DEBUG_LEVEL3
*/

#ifdef DEBUG_LEVEL0
#ifndef DEBUG_MODE
#define DEBUG_MODE
#endif
#define DEBUG0(stmts) stmts fflush(dbg_fp);
#else
#define DEBUG0(stmts)
#endif

#ifdef DEBUG_LEVEL1
#ifndef DEBUG_MODE
#define DEBUG_MODE
#endif
#define DEBUG1(stmts) stmts fflush(dbg_fp);
#else
#define DEBUG1(stmts)
#endif

#ifdef DEBUG_LEVEL2
#ifndef DEBUG_MODE
#define DEBUG_MODE
#endif
#define DEBUG2(stmts) stmts fflush(dbg_fp);
#else
#define DEBUG2(stmts)
#endif

#ifdef DEBUG_LEVEL3
#ifndef DEBUG_MODE
#define DEBUG_MODE
#endif
#define DEBUG3(stmts) stmts fflush(dbg_fp);
#else
#define DEBUG3(stmts)
#endif

#define EMPTY     1
#define NOT_EMPTY 0

#define _SUCCESS    1
#define _FAIL   0

#define TRUE    1
#define FALSE   0

#define NONE 0
#define NOT_FOUND -1

#define IDLE 0
#define BUSY 1

/***********************************************************
 *                   DEBUG VARIABLES                       *
 ***********************************************************/

/* make sure your appl. opens dbg_fp before using it but not multiple times */
FILE *dbg_fp;

/***********************************************************
 *            Termination Detection Variable               *
 ***********************************************************/
upc_lock_t *cb_lock;
strict shared int cb_count;
strict shared int cb_cancel;
strict shared int cb_done;

/***********************************************************
 *              Locality Aware Optimization                *
 ***********************************************************/

int i_am_a_local_root;		// current thread is the local root of the node
int num_local_threads;			// same as cores_per_node
int *local_thread_list;

/***********************************************************
 *             UPC Distributed Hash Map                    *
 ***********************************************************/

#define TASK_MAP_NUM_BUCKETS  1024*1024
#define TASK_MAP_NUM_ENTRIES  16*1024*1024  

typedef struct entry entry_t;
struct entry {
	uint64_t key;
	void *val;				/* points to the value stored at val_buf array */
	entry_t *next;				/* actual next entry */
	entry_t *next_free;		/* next free entry */
};

struct upc_map {
	char name[32];
	int num_buckets;				/* total number of buckets */
	int num_entries;				/* total number of entries */
	int val_size;						/* size of value */
	int entry_cnt;					/* total number of entries used */
	entry_t *buckets;				/* bucket array */
	entry_t *entries;				/* entry array */
	entry_t *free_entries;	/* free entry list */
	char *values;						/* value array */
};
typedef struct upc_map upc_map_t;

/***********************************************************
 *            UPC Task Queue Data Structures               *
 ***********************************************************/
typedef uint64_t task_handle;

typedef struct task task_header_t;
struct task {
	uint64_t taskID;			// the task ID
	uint64_t syncID;			// the sync ID (-1: if it is an independent task )
	uint64_t child_syncID;	// the sync ID (-1: if it is an independent task ) (X)
	uint64_t prev_child_syncID;	// the sync ID (-1: if it is an independent task ) (X)
	int funcID;						// the task function ID
	int sync_tcnt;			// the number tasks created in the previous phase (X: move to task_call_stack)
	int sync_phase;			// reset when a task runs (taskq_execute/inline) and increment at taskq_fence (X: move to task_call_stack)
	int owner;					// UPC thread that created this task
	//int inlined;				// True if the current task is inlined
	//int done;					// True if the current and sub-tree tasks are done (used for inlining)
	void (*func)(void *, void *);	// function pointer mapped by funcID	(valid only after localization if stolen)
	int hasCont;				// set to 1 if this task has a continuation task call in its task function body (X: move to func_type)
	void *iptr;					// input data pointer
	void *optr;					// output data pointer (it is NULL if sync_node exists)
	int	out_size;				// size of the output value	(it is redundant with func_type->osize (X)
	int	out_offs;				// an offset from the start of the parameter (X)
	task_header_t *next;	// connects all the tasks in the same sync phase (for wtaskq) (X)
};

typedef struct task_dummy task_dummy_t;
struct task_dymmy {
	uint64_t taskID;			// the task ID
	uint64_t syncID;			// the sync ID (-1: if it is an independent task )
	uint64_t child_syncID;	// the sync ID (-1: if it is an independent task ) (X)
	uint64_t prev_child_syncID;	// the sync ID (-1: if it is an independent task ) (X)
	int funcID;						// the task function ID
	int sync_tcnt;			// the number tasks created in the previous phase (X: move to task_call_stack)
	int sync_phase;			// reset when a task runs (taskq_execute/inline) and increment at taskq_fence (X: move to task_call_stack)
	int owner;					// UPC thread that created this task
	//int inlined;				// True if the current task is inlined
	//int done;					// True if the current and sub-tree tasks are done (used for inlining)
	void (*func)(void *, void *);	// function pointer mapped by funcID	(valid only after localization if stolen)
	int hasCont;				// set to 1 if this task has a continuation task call in its task function body (X: move to func_type)
	void *iptr;					// input data pointer
	void *optr;					// output data pointer (it is NULL if sync_node exists)
	int	out_size;				// size of the output value	(it is redundant with func_type->osize (X)
	int	out_offs;				// an offset from the start of the parameter (X)
	task_header_t *next;	// connects all the tasks in the same sync phase (for wtaskq) (X)
	int dummy;
};

/*
typedef struct _wait_task_node wait_task_node_t;
struct _wait_task_node {
	task_header_t *wait_task;
	wait_task_node_t *next;
};
*/

typedef struct _sync_node sync_node_t;
struct _sync_node {
	int 					ref_count;
	task_header_t *wait_task;
	//wait_task_node_t *wait_task_list;
};

typedef struct _func_type func_type_t;
struct _func_type {
	void  (*funcP)(void *, void *);	// task function address in this thread
	int		no_output_dep;		// optional: 1 if the task function has no output dependency
	int 	iSize;	// size of the input data
	int 	oSize;	// size of the output data
};

typedef struct task_msg task_msg_header_t;
struct task_msg {
  uint64_t syncID;			// INVARIABLE 
  int out_sz;			// INVARIABLE
  int offset;			// INVARIABLE
	void *optr;
//	int data;     	// variable-size user-defined output value of the task function
};

/* msgbuf stores all task completion messages sent from all other UPC threads */
struct task_msgbuf {
	int size;							/* the size of the msgbuf for each UPC thread */
//	shared int *msg_cnt;	/* number of messages in the local msgbuf */
	upc_lock_t *lock;			/* one lock for the whole sbuf in this msgbuf */
	char *sbuf_lp;
	shared char *sbuf;		/* shared local buffer from upc_alloc(msg_cnt*taskq->msg_sz) */
};
typedef struct task_msgbuf task_msgbuf_t;
typedef task_msgbuf_t *task_msgbuf_lp;
typedef shared task_msgbuf_t *task_msgbuf_sp;

/* distributed ready queue: local portion of the global stack */
struct rtaskq {
	int size;					// the local size of the global stack (note: this is a sing
	int scnt;					// the number of tasks stored in the shared task queue 
	int pcnt;					// the number of tasks stored in the local task queue 
	int ptop;					// the top index of the local queue 
	int pbot;					// the bottom pointer to the private part of the queue 
	int tail;					// the tail pointer to the shared part of the queue 
	int prev_ptop;		// the first task of the previous sync phase; sync_top==ptop outside taskq_execute
	int nAcquire;			// stats : number of Acquired tasks 
	int nRelease;			// stats : number of Released tasks 
	int failed_victim_probe;				// the number of find_victim probing
	int failed_steal_attempt;				// the number of steal failure
	int nSteal;				// stats : number of successful stealing from other theads
	int nTaskSteal;		// stats : number of tasks stealed from other theads
	int nWakeUp;			// stats: number of times cbarrier is cancelled in upc_task_isEmpty 
	int nFail;				// stats : number of Failed task steals
	int nPut;					// stats: number of tasks produced by this UPC thread 
	int nGet;					// stats: number of tasks consumed by this UPC thread 
	shared int *nProduced;	// stats: a buffer for reducing nPut values 
	shared int *nConsumed;	// stats: a buffer for reducing nGet values 
	upc_lock_t *lock;
	char *gbuf_lp;
	shared char *gbuf;	// globally shared task queue pointer 
};
typedef struct rtaskq rtaskq_t;

shared int msgbuf_cnt[THREADS];

/* upper level: global stack */
typedef struct upc_taskq taskq_t;
typedef taskq_t *taskq_handle;
struct upc_taskq {
	int taskqID;										// task queue identifier
	int task_sz;										// task size stored in the task queue: sizeof(task_t) + max_isize
	int max_iSize;									// user input data size stored in the task queue 
	int msg_sz;											// the size of the msg for task completion notification 
	uint64_t taskID;								// the globally unique but locally updated taskID 
	uint64_t syncID;								// the local var; inc at taskq_fence/taskq_wait/taskq_execute
	int threshold;									// the # of tasks to trigger a move from private to shared regions
	int steal_size;									// the number of tasks per steal 
	int chunk_size;									// the number of tasks per release/acquire
	int total_steal_trial;					// the # of steal attempt per taskq_steal 
	int preset_steal_trial;					// the # of steal attempt whose victim list is pre-calculated
	int runtime_cutoff;							// positive threshold value indicates enabled, otherwise diabled
	//uint64_t prev_syncID;					// the last syncID value when it is used for sync_node_map insert
	int numTaskFuncs;								// number of task functions
	int cores_per_node;							// number of CPU cores per node 
	func_type_t *task_func_list;		// a list of task func types
	int steal_mode;									// FULL_RAND_STEAL, LAWS, INTRA_NODE_ONLY
	int chunk_mode;									// FIXED, HALF_SIZE, ADAPTIVE
	int num_hierarchy;							// number of memory hierarchy levels
	int *victim_list;								// list of victim threads
	int *nprocs_per_level;					// LAWS: number of processors at each level
	int first_steal_from_root;      // steal the first work from the root UPC thread
	int task_stack_max;							// max # of re-entrant tasks
	int task_stack_top;							// # of re-entrant tasks: inc and dec at taskq_execute 
	task_header_t *task_get_buf;		// pre-allocated temp. buffer for ??? of size (task_sz)
	task_header_t *task_put_buf;		// pre-allocated temp. buffer for taskq_put of size (task_sz)
	char *task_out_buf;							// pre-allocated temp. buffer for output value of size max_out_sz
	task_msg_header_t *task_msg_buf;	// pre-allocated temp. buffer of size taskq->msg_sz
	upc_map_t *wait_task_map;				// local map for wait_task queue 
	upc_map_t *dep_graph_map;				// local map to dependency graph 
	upc_map_t *sync_node_map;				// local map for sync_node map 
	rtaskq_t *rtaskq_lp;						// local queue for local rtaskq 
	task_msgbuf_t *msgbuf_lp;				// local buffer for local msgbuf 
	shared rtaskq_t *rtaskq_sp;			// shared pointer to the distributed ready task queue
	shared task_msgbuf_t *msgbuf_sp;	// shared pointer to the distributed msgbuf
};

/***********************************************************
 *                 UPC Map Functions                       *
 ***********************************************************/
upc_map_t *upc_map_alloc(const char *, int, int, int);
void *upc_map_put(upc_map_t *, uint64_t, void *);
void *upc_map_get(upc_map_t *, uint64_t);
void *upc_map_remove(upc_map_t *, uint64_t);
void upc_map_free(upc_map_t *);
void upc_map_display(upc_map_t *);
void upc_map_stats(upc_map_t *);
void upc_map_snapshot(upc_map_t *, int);
void *upc_map_get_one(upc_map_t *);

/***********************************************************
 *          UPC Task Queue Internal Functions              *
 ***********************************************************/
void taskq_wakeup();
task_handle taskq_notify(taskq_handle, task_handle, void *);
void taskq_check(taskq_handle);
void taskq_local_reset(taskq_handle);
task_header_t *taskq_get(taskq_handle);
//int  taskq_insert(taskq_handle, task_header_t *, func_type_t *);  /* temporarily moved to upc_task.h */

void rtaskq_push(taskq_t *, task_header_t *, int);
int  taskq_acquire(taskq_t *);
int  taskq_release(taskq_t *);
int  taskq_find_victim(taskq_handle, int, int);
int  taskq_local_depth(taskq_handle);

int  taskq_process_msgbuf(taskq_handle);
int  taskq_steal_engine(taskq_handle, int, int);
void taskq_set_notAllEmpty(void);

void func_check(taskq_handle, uint64_t, char *);

#endif
