/***********************************************************
 *                                                         *
 *           UPC TASK PARALLEL LIBRARY HEADER              *
 *                                                         *
 ***********************************************************/

#ifndef _UPC_TASK_H
#define	_UPC_TASK_H

#include <upc.h>
#include <string.h>
#include <stdint.h>       /* data type - uintptr_t */
#include <stdarg.h>       /* variable length argument list */
#include "taskq.h"

/***********************************************************
 *                  #define VARIABLES                      *
 ***********************************************************/

//#define RUNTIME_CUTOFF

// taskq->steal_mode
#define FULL_RAND_STEAL 1
#define SAME_NODE_STEAL 2
#define HIER_WORK_STEAL 3

// taskq->steal_chunk_mode
#define FIXED_CHUNK			1	// fixed chunk_size for all 
#define HALF_CHUNK			2	// half-of-scnt for all
#define ADAPTIVE_CHUNK	3	// fixed chunk_size for intra-node, half-of-scnt for inter-node 

//#define HOPPER

/***********************************************************
1. STEAL_INTRA_NODE_ONLY		// does not steal from remote nodes
   define LAWS and LAWS_RAND 
   call taskq_set_steal_trial(taskq, 12, 12); and taskq_set_laws_level(taskq, 1, 8);

2. RANDOM
   do not define LAWS and LAWS_RAND
   call taskq_set_steal_trial(taskq, 12, 12); and DO NOT call taskq_set_laws_level();

3. LAWS
   define LAWS and LAWS_RAND 
   call taskq_set_steal_trial(taskq, 8, 12); and taskq_set_laws_level(taskq, 1, 8);
   or 
   call taskq_set_steal_trial(taskq, 12, 16); and taskq_set_laws_level(taskq, 2, 8, 12);

4. LOCAL ROOT

 ***********************************************************/

//#define LOCAL_ROOT			// only one thread per node will perform communication 

//#define STEAL_RAND 1 		// random victim selection 
//#define STEAL_NEXT 2		// MYTHREAD+1 victim selection

#ifdef RUNTIME_CUTOFF

#define READYQ_SZ     1024
#define TASK_STACK_SZ 32768    
#define WAITQ_BUCKETS 4096      
#define WAITQ_ENTRIES 16384    
#define MSGBUF_SZ     1024

#else

#define READYQ_SZ     250000
#define TASK_STACK_SZ 16384    
#define WAITQ_BUCKETS 4096      
#define WAITQ_ENTRIES 16384    
#define MSGBUF_SZ     1024

#endif

#define DEFAULT_TASKQ_STEAL_SIZE      2
#define DEFAULT_TASKQ_CHUNK_SIZE      4
#define DEFAULT_TASKQ_THRESHOLD       4

/***********************************************************
 *                                                         *
 *  functions                                              *
 *                                                         *
 ***********************************************************/

/* task queue allocation and free */
taskq_handle taskq_all_alloc(int, ...);
void taskq_all_free(taskq_handle);
void taskq_finalize();

/* task queue data functions */
int taskq_put(taskq_handle, void *func, ...);	
int taskq_put_continuation(taskq_handle, void *func, ...);	
int taskq_execute(taskq_handle);	
int taskq_steal(taskq_handle);
int taskq_isEmpty(taskq_handle);			
int taskq_all_isEmpty(taskq_handle);			
void taskq_all_run(taskq_handle);			

/* task queue synchronization */
void taskq_wait(taskq_handle);
void taskq_fence(taskq_handle);

/* task queue synchronization */
void taskq_set_steal_size(taskq_handle, int); 		
void taskq_set_steal_trial(taskq_handle, int, int); 		
void taskq_set_chunk_size(taskq_handle, int); 		
void taskq_set_threshold(taskq_handle, int); 		
void taskq_set_laws_level(taskq_handle, int, ...); 		
void taskq_set_first_steal_from_root(taskq_handle);
void taskq_set_independent_task(taskq_handle, void *);
void taskq_set_runtime_cutoff(taskq_handle, int);
void taskq_set_steal_mode(taskq_t *, int, int, ...);
void taskq_set_chunk_mode(taskq_t *, int);

/* task queue debugging */
void taskq_stats(taskq_handle);
void taskq_snapshot(taskq_handle);
void taskq_remote_stats(taskq_handle, int);
void taskq_remote_snapshot(taskq_handle, int);
void taskq_error_exit(const char *);

/* utilities */
int ilog2(int);
void print_stack_indent(void);

/* optimization */
int  taskq_insert(taskq_handle, task_header_t *, func_type_t *);

#endif
